import logging
import os
import unittest
from contextlib import contextmanager
from shutil import copyfile
from tempfile import TemporaryDirectory

import cherrypy
from cmapi_server import helpers
from cmapi_server.constants import CMAPI_CONF_PATH
from cmapi_server.controllers.dispatcher import dispatcher, jsonify_error
from cmapi_server.logging_management import config_cmapi_server_logging
from cmapi_server.managers.application import (
    AppStatefulConfig, StatefulConfigModel, StatefulFlagsModel
)
from cmapi_server.managers.process import MCSProcessManager
from cmapi_server.managers.certificate import CertificateManager


TEST_API_KEY = 'somekey123'
MCS_CONFIG_FILEPATH = '/etc/columnstore/Columnstore.xml'
# Use absolute paths relative to this test package so tests work regardless of CWD
_TEST_DIR = os.path.dirname(__file__)
COPY_MCS_CONFIG_FILEPATH = os.path.join(_TEST_DIR, 'original_Columnstore.xml')
TEST_MCS_CONFIG_FILEPATH = os.path.join(_TEST_DIR, 'CS-config-test.xml')
# TODO:
#       - rename after fix in all places
#       - fix path to abs
mcs_config_filename = os.path.join(_TEST_DIR, 'CS-config-test.xml')
tmp_mcs_config_filename = os.path.join(_TEST_DIR, 'tmp.xml')
cmapi_config_filename = os.path.join(os.path.dirname(_TEST_DIR), 'cmapi_server.conf')
tmp_cmapi_config_filename = os.path.join(_TEST_DIR, 'tmp.conf')
single_node_xml = os.path.join(_TEST_DIR, '..', 'SingleNode.xml')
# constants for process dispatchers
DDL_SERVICE = 'mcs-ddlproc'
CONTROLLERNODE_SERVICE = 'mcs-controllernode.service'
UNKNOWN_SERVICE = 'unknown_service'
SYSTEMCTL = 'systemctl'


logging.basicConfig(level=logging.DEBUG)


def run_detect_processes():
    cfg_parser = helpers.get_config_parser(CMAPI_CONF_PATH)
    d_name, d_path = helpers.get_dispatcher_name_and_path(cfg_parser)
    MCSProcessManager.detect(d_name, d_path)


@contextmanager
def run_server():
    CertificateManager.create_self_signed_certificate_if_not_exist()

    cherrypy.engine.start()
    cherrypy.engine.wait(cherrypy.engine.states.STARTED)
    run_detect_processes()  #TODO: Move cause slow down each test for 5s
    yield

    cherrypy.engine.exit()
    cherrypy.engine.block()


class BaseServerTestCase(unittest.TestCase):
    HEADERS = {'x-api-key': TEST_API_KEY}
    NO_AUTH_HEADERS = {'x-api-key': None}
    TEST_PARAMS = (
        ('auth ok', HEADERS, 200),
        ('no auth', NO_AUTH_HEADERS, 401)
    )

    def run(self, result=None):
        with TemporaryDirectory() as tmp_dir:
            self.tmp_dir = tmp_dir
            self.cmapi_config_filename = os.path.join(
                tmp_dir, 'tmp_cmapi_config.conf'
            )
            self.mcs_config_filename = os.path.join(
                tmp_dir, 'tmp_mcs_config.xml'
            )
            copyfile(cmapi_config_filename, self.cmapi_config_filename)
            copyfile(TEST_MCS_CONFIG_FILEPATH, self.mcs_config_filename)
            config_cmapi_server_logging()
            self.app = cherrypy.tree.mount(
                root=None, config=self.cmapi_config_filename
            )
            self.app.config.update({
                '/': {
                    'request.dispatch': dispatcher,
                    'error_page.default': jsonify_error,
                },
                'config': {
                    'path': self.cmapi_config_filename,
                },
                'Authentication' : self.HEADERS
            })
            cherrypy.config.update(self.cmapi_config_filename)

            with run_server():
                return super().run(result=result)


class BaseNodeManipTestCase(unittest.TestCase):
    NEW_NODE_NAME = 'node.hostname'

    def setUp(self):
        self.tmp_files = []
        copyfile(TEST_MCS_CONFIG_FILEPATH, tmp_mcs_config_filename)

    def tearDown(self):
        for tmp_file in self.tmp_files:
            if os.path.exists(tmp_file):
                os.remove(tmp_file)
        if os.path.exists(tmp_mcs_config_filename):
            os.remove(tmp_mcs_config_filename)

    def _set_shared_storage(self, target_value: bool) -> bool:
        """Set shared_storage_on flag to target_value, return original value.

        If the current value already equals target_value, no update is applied.
        """
        current_cfg = AppStatefulConfig.get_config_copy()
        original_value = current_cfg.flags.shared_storage_on
        if original_value != target_value:
            new_cfg = StatefulConfigModel(
                version=current_cfg.version.next_seq(),
                flags=StatefulFlagsModel(shared_storage_on=target_value),
            )
            AppStatefulConfig.apply_update(new_cfg)
        return original_value


class BaseProcessDispatcherCase(unittest.TestCase):
    node_started = None

    @classmethod
    def setUpClass(cls) -> None:
        run_detect_processes()
        cls.node_started = MCSProcessManager.get_running_mcs_procs() != 0
        return super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        if (MCSProcessManager.get_running_mcs_procs() !=0) == cls.node_started:
            return super().tearDownClass()
        if cls.node_started:
            MCSProcessManager.start_node(is_primary=True, use_sudo=False)
        else:
            MCSProcessManager.stop_node(is_primary=True, use_sudo=False)
        return super().tearDownClass()

    def setUp(self) -> None:
        if MCSProcessManager.process_dispatcher.is_service_running(
            CONTROLLERNODE_SERVICE
        ):
            self.controller_node_cmd = 'start'
        else:
            self.controller_node_cmd = 'stop'
        # prevent to get 'start-limit-hit' systemd error, see MCOL-5186
        os.system(f'{SYSTEMCTL} reset-failed')
        return super().setUp()

    def tearDown(self) -> None:
        os.system(
            f'{SYSTEMCTL} {self.controller_node_cmd} {CONTROLLERNODE_SERVICE}'
        )
        return super().tearDown()



"""
This script stress tests deadlock detection.

Usage: rocksdb_deadlock_stress.py user host port db_name table_name
       num_iters num_threads
"""
import cStringIO
import hashlib
import MySQLdb
from MySQLdb.constants import ER
import os
import random
import signal
import sys
import threading
import time
import string
import traceback

def is_deadlock_error(exc):
    error_code = exc.args[0]
    return (error_code == MySQLdb.constants.ER.LOCK_DEADLOCK)

def get_query(table_name, idx):
  # Let's assume that even indexes will always be acquireable, to make
  # deadlock detection more interesting.
  if idx % 2 == 0:
    return """SELECT * from %s WHERE a = %d LOCK IN SHARE MODE""" % (table_name, idx)
  else:
    r = random.randint(1, 3);
    if r == 1:
      return """SELECT * from %s WHERE a = %d FOR UPDATE""" % (table_name, idx)
    elif r == 2:
      return """INSERT INTO %s VALUES (%d, 1)
                ON DUPLICATE KEY UPDATE b=b+1""" % (table_name, idx)
    else:
      return """DELETE from %s WHERE a = %d""" % (table_name, idx)

class Worker(threading.Thread):
  def __init__(self, con, table_name, num_iters):
    threading.Thread.__init__(self)
    self.con = con
    self.table_name = table_name
    self.num_iters = num_iters
    self.exception = None
    self.start()
  def run(self):
    try:
      self.runme()
    except Exception, e:
      self.exception = traceback.format_exc()
  def runme(self):
    cur = self.con.cursor()
    for x in xrange(self.num_iters):
      try:
        for i in random.sample(xrange(100), 10):
          cur.execute(get_query(self.table_name, i))
        self.con.commit()
      except MySQLdb.OperationalError, e:
        self.con.rollback()
        cur = self.con.cursor()
        if not is_deadlock_error(e):
          raise e

if __name__ == '__main__':
  if len(sys.argv) != 8:
    print "Usage: rocksdb_deadlock_stress.py user host port db_name " \
          "table_name num_iters num_threads"
    sys.exit(1)

  user = sys.argv[1]
  host = sys.argv[2]
  port = int(sys.argv[3])
  db = sys.argv[4]
  table_name = sys.argv[5]
  num_iters = int(sys.argv[6])
  num_workers = int(sys.argv[7])

  worker_failed = False
  workers = []
  for i in xrange(num_workers):
    w = Worker(
      MySQLdb.connect(user=user, host=host, port=port, db=db), table_name,
      num_iters)
    workers.append(w)

  for w in workers:
    w.join()
    if w.exception:
      print "Worker hit an exception:\n%s\n" % w.exception
      worker_failed = True

  if worker_failed:
    sys.exit(1)



import logging
import os
import unittest
from pathlib import Path
from shutil import rmtree

from cmapi_server.constants import MCS_MODULE_FILE_PATH
from mcs_node_control.models.node_status import NodeStatus


logging.basicConfig(level='DEBUG')


class NodeStatusTest(unittest.TestCase):
    def test_dbrm_cluster_mode(self):
        node_status = NodeStatus()
        # use subprocess.run to capture stdout
        os.system('/usr/bin/dbrmctl readwrite')
        self.assertEqual(node_status.get_cluster_mode(), 'readwrite')
        os.system('/usr/bin/dbrmctl readonly')
        self.assertEqual(node_status.get_cluster_mode(), 'readonly')
        # kill controllernode and test it

    def test_dbrm_status(self):
        node_status = NodeStatus()
        self.assertEqual(node_status.get_dbrm_status(), 'master')

    def test_dbroots(self):
        try:
            node_status = NodeStatus()
            dbroot_ids = [1, 2, 3]
            path = '/tmp/dbroots/'
            for e in dbroot_ids:
                p = Path(path + 'data' + str(e))
                p.mkdir(parents = True, exist_ok = True)
            for e in node_status.get_dbroots(path=path):
                self.assertEqual(e in dbroot_ids, True)
        except AssertionError as e:
            rmtree(path)
            raise e

    def test_module_id(self):
        node_status = NodeStatus()
        module_file = Path(MCS_MODULE_FILE_PATH)
        examplar_id = int(module_file.read_text()[2:])
        self.assertEqual(examplar_id, node_status.get_module_id())


if __name__ == '__main__':
    unittest.main()



import logging
import os
import re
import subprocess

import requests

from cmapi_server.constants import (
    ES_REPO, ES_REPO_PRIORITY_PREFS, ES_TOKEN_VERIFY_URL, ES_VERIFY_URL, MDB_GPG_KEY_URL,
    MDB_LATEST_RELEASES_URL, MDB_LATEST_TESTED_MAJOR, REQUEST_TIMEOUT,
)
from cmapi_server.exceptions import CMAPIBasicError
from cmapi_server.managers.upgrade.utils import ComparableVersion
from cmapi_server.process_dispatchers.base import BaseDispatcher


class MariaDBESRepoManager:
    def __init__(
            self, token: str, arch: str, os_type: str, os_version: str,
            mariadb_version: str = 'latest',
    ):
        self.token = token
        self.arch = arch
        self.os_type = os_type
        self.os_version = os_version
        if mariadb_version == 'latest':
            self.mariadb_version = self.get_latest_tested_mdb_version()
        else:
            self.mariadb_version = mariadb_version

    def _import_mariadb_keyring(self):
        """
        Download and place the MariaDB keyring into /etc/apt/trusted.gpg.d.
        """
        key_url = 'https://supplychain.mariadb.com/mariadb-keyring-2019.gpg'
        keyring_path = '/etc/apt/trusted.gpg.d/mariadb-keyring-2019.gpg'

        try:
            # Download the keyring file
            response = requests.get(key_url)
            response.raise_for_status()

            # Write the keyring file to the specified path
            with open(keyring_path, 'wb') as key_file:
                key_file.write(response.content)

            # Set permissions to 644
            os.chmod(keyring_path, 0o644)
            logging.debug(
                f'Keyring successfully downloaded and placed at {keyring_path}'
            )
        except requests.RequestException as exc:
            raise CMAPIBasicError(
                f'Failed to download keyring from {key_url}: {exc}'
            )
        except OSError as exc:
            raise CMAPIBasicError(
                f'Failed to write keyring to {keyring_path}: {exc}'
            )

    def check_mdb_version_exists(self):
        """Check if passed MDB version exists in the repo.

        :raises CMAPIBasicError: unsupported OS type
        :raises CMAPIBasicError: wrong MDB version passed
        :raises CMAPIBasicError: some other request/response errors
        """
        verify_url: str = ''
        if self.os_type in ['ubuntu', 'debian']:
            # get only two first numbers from version to build repo link
            verify_url = ES_VERIFY_URL.deb.format(
                token=self.token,
                mdb_version=self.mariadb_version,
                os_version=self.os_version
            )
        elif self.os_type in ['centos', 'rhel', 'rocky']:
            verify_url = ES_VERIFY_URL.rhel.format(
                token=self.token,
                mdb_version=self.mariadb_version,
                os_major_version=self.os_version.split('.', maxsplit=1)[0],
                arch=self.arch
            )
        else:
            raise CMAPIBasicError(f'Unsupported OS type: {self.os_type}')
        try:
            # Download the keyring file
            response = requests.get(verify_url, timeout=REQUEST_TIMEOUT)
            if response.status_code in (403, 404):
                raise CMAPIBasicError(
                    'MariaDB Enterprise Server version '
                    f'{self.mariadb_version} is not working for your OS '
                    'version or OS type.\nPlease verify that it is correct.\n '
                    'Not all releases of MariaDB are available on all '
                    'distributions.'
                )
            elif response.ok:
                logging.debug(
                    'MariaDB Enterprise Server version '
                    f'{self.mariadb_version} is valid.'
                )
            else:
                response.raise_for_status()
        except requests.RequestException:
            raise CMAPIBasicError(
                'Failed to check MDB version exists from '
                f'{verify_url}'
            )

    @staticmethod
    def verify_token(token: str):
        """Verify ES token.

        :param token: es token to verify
        :type token: str
        :raises CMAPIBasicError: Invalid token format
        :raises CMAPIBasicError: Invalid token
        :raises CMAPIBasicError: Other request errors
        """
        # Check token format UUID
        valid_format = re.fullmatch(
            r'[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}',
            token
        )
        if not valid_format:
            err_message = (
                f'Invalid token format: "{token}". The token should be '
                'of the form ########-####-####-####-############'
            )
            raise CMAPIBasicError(err_message)

        verify_url = ES_TOKEN_VERIFY_URL.format(token=token)

        try:
            response = requests.head(
                verify_url, allow_redirects=True, timeout=REQUEST_TIMEOUT
            )
            if response.status_code in (403, 404):
                raise CMAPIBasicError(
                    'Invalid token. Please verify that it is correct.'
                )
            elif response.ok:
                logging.debug('MariaDB ES Token is valid.')
            else:
                response.raise_for_status()
        except requests.RequestException:
            raise CMAPIBasicError(
                'Problem encountered while trying to verify ES token.'
            )

    @classmethod
    def get_latest_tested_mdb_version(cls) -> str:
        """Get latest tested MDB version from repo.

        :raises CMAPIBasicError: no latest version matched with latest tested
        :raises CMAPIBasicError: if request error
        :return: latest MDB version matched with latest tested major
        """
        try:
            # Download the keyring file
            response = requests.get(
                MDB_LATEST_RELEASES_URL, timeout=REQUEST_TIMEOUT
            )
            response.raise_for_status()
            latest_version_nums = [
                ver
                for ver in response.text.split(' ')
                if MDB_LATEST_TESTED_MAJOR in ver
            ]
            if not latest_version_nums:
                raise CMAPIBasicError(
                    'Failed to get latest MDB version number matched latest '
                    f'tested major {MDB_LATEST_TESTED_MAJOR}'
            )
            latest_version_num = sorted(latest_version_nums, reverse=True)[0]
            logging.debug(
                'Succesfully got latest MDB version number: '
                f'{latest_version_num}'
            )
        except requests.RequestException as exc:
            raise CMAPIBasicError(
                'Failed to get latest MDB version numbers from '
                f'{MDB_LATEST_RELEASES_URL}'
            )
        return latest_version_num

    @classmethod
    def get_ver_of(cls, package_name: str, os_type: str,) -> str:
        """Get version of package in a repo.

        :param package_name: name of package to get version
        :type package_name: str
        :param os_type: os name
        :type os_type: str
        :raises CMAPIBasicError: if os type isn't supported
        :raises CMAPIBasicError: if failed getting package info
        :raises CMAPIBasicError: if couldn't find any package with given name
        :return: latest available package version
        :rtype: str
        """
        latest_version: str = ''
        cmd: str = ''
        if os_type in ['ubuntu', 'debian']:
            cmd = f'apt show {package_name}'
        elif os_type in ['centos', 'rhel', 'rocky']:
            cmd = f'yum info --showduplicates --available {package_name}'
        else:
            raise CMAPIBasicError(f'Unsupported OS type: {os_type}')

        success, result = BaseDispatcher.exec_command(cmd)
        if not success:
            message = (
                f'Failed to get {package_name} package information using '
                f'command {cmd} with result: {result}'
            )
            logging.error(message)
            raise CMAPIBasicError(message)

        matches = re.findall(r'\bVersion\s*:\s+(\S+)', result)
        if matches:
            latest_version = max(matches, key=ComparableVersion)
        else:
            raise CMAPIBasicError(
                'Could not find any version for package with name '
                f'{package_name}'
            )
        return latest_version

    def check_repo(self):
        """Check repo installed and have needed version of MDB.

        :raises CMAPIBasicError: if unsupported os type detected
        :raises CMAPIBasicError: could not find package matching the version
        """
        pkg_ver: str = ''
        mdb_pkg_mgr_version = self.get_ver_of('mariadb-server', self.os_type)
        if self.os_type in ['ubuntu', 'debian']:
            # for deb packages it's just a part of version
            # eg: 10.6.22.18 and 1:10.6.22.18+maria~ubu2204
            pkg_ver = self.mariadb_version.replace('-', '.')
        elif self.os_type in ['centos', 'rhel', 'rocky']:
            # in rhel distros it's full version
            pkg_ver = self.mariadb_version.replace('-', '_')
        else:
            raise CMAPIBasicError(f'Unsupported OS type: {self.os_type}')
        if pkg_ver not in mdb_pkg_mgr_version:
            raise CMAPIBasicError(
                'Could not find mariadb-server package matched with version '
                f'{pkg_ver}'
            )

    def setup_repo(self):
        """Set up the MariaDB Enterprise repository based on OS type."""
        self.check_mdb_version_exists()
        if self.os_type in ['ubuntu', 'debian']:
            # get only two first numbers from version to build repo link
            repo_data = ES_REPO.deb.format(
                token=self.token,
                mdb_version=self.mariadb_version,
                os_version=self.os_version
            )
            repo_file = '/etc/apt/sources.list.d/mariadb.list'
            with open(repo_file, 'w', encoding='utf-8') as f:
                f.write(repo_data)
            # Set permissions to 640
            os.chmod(repo_file, 0o640)

            pref_file = '/etc/apt/preferences'
            with open(pref_file, 'w', encoding='utf-8') as f:
                f.write(ES_REPO_PRIORITY_PREFS)

            self._import_mariadb_keyring()
            subprocess.run(['apt-get', 'update'], check=True)
        elif self.os_type in ['centos', 'rhel', 'rocky']:
            repo_data = ES_REPO.rhel.format(
                token=self.token,
                mdb_version=self.mariadb_version,
                os_major_version=self.os_version.split('.', maxsplit=1)[0],
                arch=self.arch,
                gpg_key_url=MDB_GPG_KEY_URL
            )
            repo_file = '/etc/yum.repos.d/mariadb.repo'
            with open(repo_file, 'w', encoding='utf-8') as f:
                f.write(repo_data)
            subprocess.run(['rpm', '--import', MDB_GPG_KEY_URL], check=True)
        else:
            raise CMAPIBasicError(f'Unsupported OS type: {self.os_type}')
        self.check_repo()



from __future__ import annotations
import logging
from pathlib import Path

from lxml import etree

from cmapi_server.constants import (
    DEFAULT_MCS_CONF_PATH, MCS_DATA_PATH, MCS_MODULE_FILE_PATH,
)


module_logger = logging.getLogger()


def read_module_id():
    """Retrieves module ID from MCS_MODULE_FILE_PATH.

    :rtype: int : seconds
    """
    module_file = Path(MCS_MODULE_FILE_PATH)
    return int(module_file.read_text()[2:])


# TODO: Useless for now, newer called in code
#       Nodeconfig.apply_config doing this.
def set_module_id(module_id: int = 1):
    """Sets current module ID from MCS_MODULE_FILE_PATH.

    :rtype: int : seconds
    """
    module_file = Path(MCS_MODULE_FILE_PATH)
    return module_file.write_text(f'pm{module_id}\n')


def get_dbroots_list(path: str = MCS_DATA_PATH):
    """searches for services

    The method returns numeric ids of dbroots available.

    :rtype: generator of ints
    """
    func_name = 'get_dbroots_list'
    path = Path(path)
    for child in path.glob('data[1-9]*'):
        dir_list = str(child).split('/') # presume Linux only
        dbroot_id = int(''.join(list(filter(str.isdigit, dir_list[-1]))))
        module_logger.debug(f'{func_name} The node has dbroot {dbroot_id}')
        yield dbroot_id


def get_workernodes() -> dict[dict[str, int]]:
    """Get workernodes list.

    Returns a list of network address of all workernodes.
    This is an equivalent of all nodes.

    :return: workernodes dict
    :rtype: dict[dict[str, int]]
    """
    # TODO: fix in MCOL-5147, get xml path from class that will handle xml
    root = current_config_root()
    workernodes = {}
    # searches for all tags starts with DBRM_Worker, eg DBRM_Worker1
    workernodes_elements = root.xpath(
        "//*[starts-with(local-name(), 'DBRM_Worker')]"
    )
    for workernode_el in workernodes_elements:
        workernode_ip = workernode_el.find('./IPAddr').text
        if workernode_ip == '0.0.0.0':
            # skip elements with specific ip
            continue
        try:
            workernode_port = int(workernode_el.find('./Port').text)
        except (AttributeError, ValueError):
            # AttributeError for not found Port tag, so got None.text
            # ValueError for non numeric values in tag text
            module_logger.error(
                'No Port tag found or wrong Port value for tag '
                f'"{workernode_el.tag}".'
            )
            workernode_port = 8700
        workernodes[workernode_el.tag] = {
            'IPAddr': workernode_ip, 'Port': workernode_port
        }
    return workernodes


def get_dbrm_master(config_filename: str = DEFAULT_MCS_CONF_PATH) -> dict:
    """Get DBRM master ip and port.

    :param config_filename: path to xml conf, defaults to DEFAULT_MCS_CONF_PATH
    :type config_filename: str, optional
    :return: ipaddress and port of DBRM master
    :rtype: dict
    """
    # TODO: fix in MCOL-5147, get xml path from class that will handle xml
    #       Use NodeConfig class as a template?
    root = current_config_root(config_filename)
    return {
        'IPAddr': root.find("./DBRM_Controller/IPAddr").text,
        'Port': root.find("./DBRM_Controller/Port").text
    }


def current_config_root(config_filename: str = DEFAULT_MCS_CONF_PATH):
    """Retrievs current configuration

    Read the config and returns Element

    :rtype: lxml.Element
    """
    parser = etree.XMLParser(load_dtd=True)
    tree = etree.parse(config_filename, parser=parser)
    return tree.getroot()



import logging
import time
from typing import Any, Optional

import cherrypy

from mcs_node_control.models.node_config import NodeConfig
from tracing.tracer import TracerBackend, TraceSpan
from tracing.utils import swallow_exceptions

logger = logging.getLogger('tracer')
json_logger = logging.getLogger('json_trace')


class TraceparentBackend(TracerBackend):
    """Default backend that logs span lifecycle and mirrors events/status."""
    def __init__(self):
        my_addresses = list(NodeConfig().get_network_addresses())
        logger.info('My addresses: %s', my_addresses)
        json_logger.info('my_addresses', extra={'my_addresses': my_addresses})

    @swallow_exceptions
    def on_span_start(self, span: TraceSpan) -> None:
        json_logger.info('span_begin', extra=span.to_flat_dict())

    @swallow_exceptions
    def on_span_end(self, span: TraceSpan, exc: Optional[BaseException]) -> None:
        end_ns = time.time_ns()
        duration_ms = (end_ns - span.start_ns) / 1_000_000
        span.set_attribute('duration_ms', duration_ms)
        span.set_attribute('end_ns', end_ns)

        # Try to set status code if not already set
        if span.kind == 'SERVER' and 'http.status_code' not in span.attributes:
            try:
                status_str = str(cherrypy.response.status)
                code = int(status_str.split()[0])
                span.set_attribute('http.status_code', code)
            except Exception:
                pass

        json_logger.info('span_end', extra=span.to_flat_dict())

    @swallow_exceptions
    def on_span_event(self, span: TraceSpan, name: str, attrs: dict[str, Any]) -> None:
        json_logger.info('span_event', extra=span.to_flat_dict())





#!/usr/bin/python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

# amalgamate.py creates an amalgamation from a unity build.
# It can be run with either Python 2 or 3.
# An amalgamation consists of a header that includes the contents of all public
# headers and a source file that includes the contents of all source files and
# private headers.
#
# This script works by starting with the unity build file and recursively expanding
# #include directives. If the #include is found in a public include directory,
# that header is expanded into the amalgamation header.
#
# A particular header is only expanded once, so this script will
# break if there are multiple inclusions of the same header that are expected to
# expand differently. Similarly, this type of code causes issues:
#
# #ifdef FOO
#   #include "bar.h"
#   // code here
# #else
#   #include "bar.h"            // oops, doesn't get expanded
#   // different code here
# #endif
#
# The solution is to move the include out of the #ifdef.

from __future__ import print_function

import argparse
from os import path
import re
import sys

include_re = re.compile('^[ \t]*#include[ \t]+"(.*)"[ \t]*$')
included = set()
excluded = set()

def find_header(name, abs_path, include_paths):
    samedir = path.join(path.dirname(abs_path), name)
    if path.exists(samedir):
        return samedir
    for include_path in include_paths:
        include_path = path.join(include_path, name)
        if path.exists(include_path):
            return include_path
    return None

def expand_include(include_path, f, abs_path, source_out, header_out, include_paths, public_include_paths):
    if include_path in included:
        return False

    included.add(include_path)
    with open(include_path) as f:
        print('#line 1 "{}"'.format(include_path), file=source_out)
        process_file(f, include_path, source_out, header_out, include_paths, public_include_paths)
    return True

def process_file(f, abs_path, source_out, header_out, include_paths, public_include_paths):
    for (line, text) in enumerate(f):
        m = include_re.match(text)
        if m:
            filename = m.groups()[0]
            # first check private headers
            include_path = find_header(filename, abs_path, include_paths)
            if include_path:
                if include_path in excluded:
                    source_out.write(text)
                    expanded = False
                else:
                    expanded = expand_include(include_path, f, abs_path, source_out, header_out, include_paths, public_include_paths)
            else:
                # now try public headers
                include_path = find_header(filename, abs_path, public_include_paths)
                if include_path:
                    # found public header
                    expanded = False
                    if include_path in excluded:
                        source_out.write(text)
                    else:
                        expand_include(include_path, f, abs_path, header_out, None, public_include_paths, [])
                else:
                    sys.exit("unable to find {}, included in {} on line {}".format(filename, abs_path, line))

            if expanded:
                print('#line {} "{}"'.format(line+1, abs_path), file=source_out)
        elif text != "#pragma once\n":
            source_out.write(text)

def main():
    parser = argparse.ArgumentParser(description="Transform a unity build into an amalgamation")
    parser.add_argument("source", help="source file")
    parser.add_argument("-I", action="append", dest="include_paths", help="include paths for private headers")
    parser.add_argument("-i", action="append", dest="public_include_paths", help="include paths for public headers")
    parser.add_argument("-x", action="append", dest="excluded", help="excluded header files")
    parser.add_argument("-o", dest="source_out", help="output C++ file", required=True)
    parser.add_argument("-H", dest="header_out", help="output C++ header file", required=True)
    args = parser.parse_args()

    include_paths = list(map(path.abspath, args.include_paths or []))
    public_include_paths = list(map(path.abspath, args.public_include_paths or []))
    excluded.update(map(path.abspath, args.excluded or []))
    filename = args.source
    abs_path = path.abspath(filename)
    with open(filename) as f, open(args.source_out, 'w') as source_out, open(args.header_out, 'w') as header_out:
        print('#line 1 "{}"'.format(filename), file=source_out)
        print('#include "{}"'.format(header_out.name), file=source_out)
        process_file(f, abs_path, source_out, header_out, include_paths, public_include_paths)

if __name__ == "__main__":
    main()



import logging
import socket

from cmapi_server.constants import DEFAULT_MCS_CONF_PATH
from mcs_node_control.models.dbrm_socket import (
    DBRM_COMMAND_BYTES, DEFAULT_HOST, DEFAULT_PORT, DBRMSocketHandler
)
from mcs_node_control.models.node_config import NodeConfig
from mcs_node_control.models.process import Process


# TODO: why we need bitwise shift here? May be constant values?
SYSTEM_STATE_FLAGS = {
    "SS_READY":            1 << 0,  # 1
    "SS_SUSPENDED":        1 << 1,  # 2
    "SS_SUSPEND_PENDING":  1 << 2,  # 4
    "SS_SHUTDOWN_PENDING": 1 << 3,  # 8
    "SS_ROLLBACK":         1 << 4,  # 16
    "SS_FORCE":            1 << 5,  # 32
    "SS_QUERY_READY":      1 << 6,  # 64
}


module_logger = logging.getLogger()


class DBRM:
    """Class DBRM commands"""
    def __init__(
        self, root=None, config_filename: str = DEFAULT_MCS_CONF_PATH
    ):
        self.dbrm_socket = DBRMSocketHandler()
        self.root = root
        self.config_filename = config_filename

    def connect(self):
        node_config = NodeConfig()
        root = self.root or node_config.get_current_config_root(
            self.config_filename
        )
        master_conn_info = node_config.get_dbrm_conn_info(root)
        if master_conn_info is None:
            module_logger.warning(
                 'DBRB.connect: No DBRM info in the Columnstore.xml.'
            )
        dbrm_host = master_conn_info['IPAddr'] or DEFAULT_HOST
        dbrm_port = int(master_conn_info['Port']) or DEFAULT_PORT
        self.dbrm_socket.connect(dbrm_host, dbrm_port)

    def close(self):
        self.dbrm_socket.close()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        if exc_type:
            return False
        return True

    def _send_command(self, command_name, command_value=None):
        if command_name not in DBRM_COMMAND_BYTES:
            module_logger.warning(
                f'DBRM._send_command: Wrong command requested {command_name}'
            )
            return None

        module_logger.info(
            f'DBRM._send_command: Command {command_name} '
            f'was requested with value {command_value}'
        )

        self.dbrm_socket.send(command_name, command_value)
        response_value_bytes = self.dbrm_socket.receive()

        if command_name == 'readonly':
            reply = int.from_bytes(response_value_bytes, 'little')
        else:
            # get first byte, it's an error message
            err = int.from_bytes(response_value_bytes[:1], 'little')

            if err != 0:
                module_logger.warning(
                    f'DBRM._send_command: Command {command_name} '
                    'returned error on server'
                )
                raise RuntimeError(
                    f'Controller Node replied error with code {err} '
                    f'for command {command_name}'
                )

            if len(response_value_bytes) < 2:
                return None

            reply = int.from_bytes(response_value_bytes[1:], 'little')
        return reply

    def get_system_state(self):
        state = self._send_command('get_system_state')
        return [
            flag_name for flag_name, flag_value in SYSTEM_STATE_FLAGS.items()
            # TODO: looks like weird logic? Not readable.
            if flag_value & state
        ]

    def _edit_system_state(self, states: list, command: str):
        state = 0
        # TODO: why we need this? States type is list.
        #       May be str without loop inside is more appropriate here.
        if isinstance(states, str):
            states = (states,)

        for state_name in states:
            if state_name not in SYSTEM_STATE_FLAGS:
                module_logger.warning(
                    f'DBRM.{command}: Wrong system state requested: '
                    f'{state_name}'
                )
                continue
            # TODO: For that case it's same with simple addition?
            #       So why we need bitwise OR?
            state |= SYSTEM_STATE_FLAGS[state_name]

        self._send_command(command, state)

    def set_system_state(self, states: list):
        self._edit_system_state(states, 'set_system_state')

    def clear_system_state(self, states: list):
        self._edit_system_state(states, 'clear_system_state')

    @staticmethod
    def get_dbrm_status():
        """Reads DBRM status

        DBRM Block Resolution Manager operates in two modes:
            - master
            - slave

        This method returns the mode of this DBRM node
        looking for controllernode process running.

        :return: mode of this DBRM node
        :rtype: string
        """
        if Process.check_process_alive('controllernode'):
            return 'master'
        return 'slave'

    def _get_cluster_mode(self):
        """Get DBRM cluster mode for internal usage.

        Returns real DBRM cluster mode from socket response.
        """
        # state can be 1(readonly) or 0(readwrite) or exception raised
        state = self._send_command('readonly')
        if state == 1:
            return 'readonly'
        elif state == 0:
            return 'readwrite'

    def get_cluster_mode(self):
        """Get DBRM cluster mode for external usage.

        There are some kind of weird logic.
        It's requested from management.
        TODO: Here we can cause a logic error.
              E.g. set non master node to "readwrite" and
                we got a "readonly" in return value.

        :return: DBRM cluster mode
        :rtype: str
        """
        real_mode = self._get_cluster_mode()
        if self.get_dbrm_status() == 'master':
            return real_mode
        else:
            return 'readonly'

    def set_cluster_mode(self, mode):
        """Set cluster mode requested

        Connects to the DBRM master's socket and
        send a command to set cluster mode.

        :rtype: str :error or cluster mode set
        """

        if mode == 'readonly':
            command = 'set_readonly'
        elif mode == 'readwrite':
            command = 'set_readwrite'
        else:
            return ''

        _ = self._send_command(command)

        return self.get_cluster_mode()


def set_cluster_mode(
    mode: str, root=None, config_filename: str = DEFAULT_MCS_CONF_PATH
):
    """Set cluster mode requested

    Connects to the DBRM master's socket and send a command to
    set cluster mode.

    :rtype: str :error or cluster mode set
    """
    try:
        with DBRM(root, config_filename) as dbrm:
            return dbrm.set_cluster_mode(mode)
    except (ConnectionRefusedError, RuntimeError, socket.error):
        module_logger.warning(
            'Cannot establish DBRM connection.', exc_info=True
        )
        return 'readonly'



#!/usr/bin/env python3
import subprocess
import sys
import xml.etree.ElementTree as ET
import configparser

XML_CONFIG_PATH = '/etc/columnstore/Columnstore.xml'
SM_CONFIG_PATH = '/etc/columnstore/storagemanager.cnf'
REST_REQUEST_TO = 2


def get_version():
    return '0.4.0'


def get_port():
    return '8640'


if __name__ == '__main__':
    master_addr = ''
    pm_count = 0
    try:
        cs_config = ET.parse(XML_CONFIG_PATH)
        config_root = cs_config.getroot()
        master_addr = config_root.find('./DBRM_Controller/IPAddr').text
        pm_count = int(config_root.find('./SystemModuleConfig/ModuleCount3').text)
    except (FileNotFoundError, AttributeError, ValueError) as e:
        print("Exception had been raised. Continue anyway")
        print(str(e))

    storage = 'LocalStorage'
    sm_config = configparser.ConfigParser()
    files_read = len(sm_config.read(SM_CONFIG_PATH))
    if files_read == 1:
        storage = sm_config.get('ObjectStorage', 'service')

    default_addr = '127.0.0.1'
    savebrm = 'save_brm'
    is_primary = False

    # For multi-node with local storage or default installations
    if (storage.lower() != 's3' and master_addr != default_addr) or \
master_addr == default_addr:
        is_primary = True
        print('Multi-node with local-storage detected.')
    else:
        has_requests = False
        try:
            import requests
            requests.packages.urllib3.disable_warnings()
            has_requests = True
        except ImportError as e:
            print('requests Python module does not exist. \
    Please install CMAPI first.')
        if has_requests is True:
            try:
                print('Requesting for the primary node status.')
                api_version = get_version()
                api_port = get_port()
                url = "https://{}:{}/cmapi/{}/node/primary".format(default_addr, \
    api_port, api_version)
                resp = requests.get(url,
                                 verify=False,
                                 timeout=REST_REQUEST_TO)
                if (resp.status_code != 200):
                    print("Error sending GET /node/primary.")
                else:
                    is_primary = resp.json()['is_primary'] == 'True'
            except:
                print('Failed to request.')
                print(str(e))

    if is_primary is True:
        try:
            retcode = subprocess.call(savebrm, shell=True)
            if retcode < 0:
                print('{} exits with {}.'.format(savebrm, retcode))
                sys.exit(0)
        except OSError as e:
                print(str(e))
                sys.exit(0)

    sys.exit(0)



#!/usr/bin/env python3

import argparse
import fnmatch
import sys

def read_patterns(filename):
    with open(filename, 'r') as f:
        return [line.strip() for line in f if line.strip()]

def read_paths_with_comments(filename):
    paths = []
    with open(filename, 'r') as f:
        for line in f:
            raw = line.rstrip('\n')
            if not raw.strip():
                continue
            # Split into path and comment
            if '#' in raw:
                path, comment = raw.split('#', 1)
                path = path.strip()
                comment = comment.strip()
            else:
                path = raw.strip()
                comment = ''
            if path:
                paths.append((raw, path, comment))
    return paths

def print_red(msg):
    if sys.stdout.isatty():
        print("\033[31m" + msg + "\033[0m")
    else:
        print(msg)

def main():
    parser = argparse.ArgumentParser(description="Compare wildcard patterns to file paths.")
    parser.add_argument('patterns_file', help='File with wildcard patterns (one per line)')
    parser.add_argument('paths_file', help='File with file paths (one per line)')
    args = parser.parse_args()

    patterns = read_patterns(args.patterns_file)
    paths_with_comments = read_paths_with_comments(args.paths_file)

    # Track which paths are matched (by index)
    path_matched = [False] * len(paths_with_comments)

    # Collect patterns that don't match any path
    unmatched_patterns = []
    for pattern in patterns:
        matched = False
        for i, (raw, path, comment) in enumerate(paths_with_comments):
            if fnmatch.fnmatch(path, pattern):
                matched = True
                path_matched[i] = True
        if not matched:
            unmatched_patterns.append(pattern)

    if unmatched_patterns:
        print_red("The files declared in debian/mariadb-plugin-columnstore.install "
                  " are not added to CMakeLists via columnstore_* statements see cmake/ColumnstoreLibrary.cmake")
        for pattern in unmatched_patterns:
            print(f"- {pattern}")

    # Collect paths that weren't matched by any pattern
    unmatched_paths = []
    for (raw, path, comment), matched in zip(paths_with_comments, path_matched):
        if not matched:
            unmatched_paths.append((path, comment))

    if unmatched_paths:
        print_red("The files added via columnstore_* statements from cmake/ColumnstoreLibrary.cmake "
                  "are missing in debian/mariadb-plugin-columnstore.install file")
        for path, comment in unmatched_paths:
            if comment:
                print(f"- {path} {comment}")
            else:
                print(f"- {path}")

if __name__ == "__main__":
    main()



import logging

import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration

from cmapi_server import helpers
from cmapi_server.constants import CMAPI_CONF_PATH
from cmapi_server.managers.application import AppManager
from tracing.sentry_backend import SentryBackend
from tracing.tracer import get_tracer

SENTRY_ACTIVE = False

logger = logging.getLogger(__name__)

def maybe_init_sentry() -> bool:
    """Initialize Sentry from CMAPI configuration.

    Reads config and initializes Sentry only if dsn parameter is present
      in corresponding section of cmapi config.
    The initialization enables LoggingIntegration: it captures error-level logs as Sentry events
      and uses lower-level logs as breadcrumbs.

    Returns: True if Sentry is initialized, False otherwise.
    """
    global SENTRY_ACTIVE
    try:
        cfg_parser = helpers.get_config_parser(CMAPI_CONF_PATH)
        dsn = helpers.dequote(
            cfg_parser.get('Sentry', 'dsn', fallback='').strip()
        )
        if not dsn:
            return False

        environment = helpers.dequote(
            cfg_parser.get('Sentry', 'environment', fallback='development').strip()
        )
    except Exception:
        logger.exception('Failed to initialize Sentry.')
        return False

    try:
        sentry_logging = LoggingIntegration(
            level=logging.INFO,
            event_level=logging.ERROR,
        )

        # Compose release string: cmapi-version(+shortrev)
        version = AppManager.get_version()
        shortrev = (AppManager.get_git_revision() or '').strip()
        shortrev = shortrev[:12] if shortrev else ''
        release = version if not shortrev else f"{version}+{shortrev}"

        sentry_sdk.init(
            dsn=dsn,
            environment=environment,
            sample_rate=1.0,
            traces_sample_rate=1.0,
            integrations=[sentry_logging],
            release=release,
        )
        # Add tags for easier filtering in Sentry
        if shortrev:
            sentry_sdk.set_tag("git_revision", shortrev)
        sentry_sdk.set_tag("cmapi_version", version)
        SENTRY_ACTIVE = True
        # Register backend to mirror our internal spans into Sentry
        get_tracer().register_backend(SentryBackend())
        logger.info('Sentry initialized for CMAPI via config.')
    except Exception:
        logger.exception('Failed to initialize Sentry.')
        return False

    logger.info('Sentry successfully initialized.')
    return True




from integration_tests.ssh import ClusterConfig
from integration_tests.utils import assert_dict_includes, drop_timestamp

ALL_MCS_SERVICES = {"StorageManager", "workernode", "controllernode", "PrimProc", "WriteEngineServer", "DDLProc", "DMLProc"}


class TestAddNode:
    def test_add_nodes(self, cluster_config: ClusterConfig, disconnected_cluster):
        """Test adding hosts to the cluster"""
        # Add primary host
        print(f"Adding self ({cluster_config.primary}) to the cluster")
        primary_private_ip = cluster_config.primary.private_ip
        add_out = cluster_config.primary.exec_mcs(f"cluster node add --node {primary_private_ip}")
        assert add_out[0]["node_id"] == str(primary_private_ip)

        status_out = cluster_config.primary.exec_mcs("cluster status")
        assert status_out["num_nodes"] == 1
        assert primary_private_ip in status_out
        primary_status = drop_timestamp(status_out[primary_private_ip])
        assert_dict_includes(
            primary_status,
            {
                "dbrm_mode": "master",
                "cluster_mode": "readwrite",
                "dbroots": ['1'],
                "module_id": 1,
            },
        )
        services_in_status = {service_descr["name"] for service_descr in primary_status["services"]}
        assert services_in_status == ALL_MCS_SERVICES

        # Add replicas
        print(f"Adding replicas ({cluster_config.replicas}) to the cluster")
        # Checking passing more that one node to the command
        add_cmd = "cluster node add "
        for replica in cluster_config.replicas:
            add_cmd += f"--node {replica.private_ip} "
        add_out = cluster_config.primary.exec_mcs(add_cmd)
        assert len(add_out) == len(cluster_config.replicas)
        added_node_ids = {node["node_id"] for node in add_out}
        expected_node_ids = {str(replica.private_ip) for replica in cluster_config.replicas}
        assert added_node_ids == expected_node_ids

        status_out = cluster_config.primary.exec_mcs("cluster status")
        assert status_out["num_nodes"] == len(cluster_config.mcs_hosts)
        for idx, replica in enumerate(cluster_config.replicas, start=2):
            assert str(replica.private_ip) in status_out
            replica_status = drop_timestamp(status_out[str(replica.private_ip)])
            assert_dict_includes(
                replica_status,
                {
                    "dbrm_mode": "slave",
                    "cluster_mode": "readonly",
                    "dbroots": [str(idx)],
                    "module_id": idx,
                },
            )
            replica_expected_services = ALL_MCS_SERVICES - {"controllernode", "DDLProc", "DMLProc"}
            services_in_status = {service_descr["name"] for service_descr in replica_status["services"]}
            assert services_in_status == replica_expected_services


"""Customized requests.Session that automatically traces outbound HTTP calls."""
from typing import Any, Optional

import logging
import time
import requests

from tracing.tracer import get_tracer, TraceSpan

# Limit for raw JSON string preview (in characters)
_PREVIEW_MAX_CHARS = 512

logger = logging.getLogger("tracer")


class TracedSession(requests.Session):
    def request(self, method: str, url: str, *args: Any, **kwargs: Any) -> requests.Response:
        tracer = get_tracer()

        headers = kwargs.get("headers") or {}
        if headers is None:
            headers = {}
        kwargs["headers"] = headers

        span_name = f"HTTP {method} {url}"
        with tracer.start_as_current_span(span_name, kind="CLIENT") as span:
            span.set_attribute("http.method", method)
            span.set_attribute("http.url", url)
            span.set_attribute("request_is_sync", True)

            tracer.inject_outbound_headers(headers)
            try:
                response = super().request(method, url, *args, **kwargs)
            except Exception as exc:
                span.set_status("ERROR", str(exc))
                raise
            else:
                span.set_attribute("http.status_code", response.status_code)
                _record_outbound_json_preview(response, span)
                return response
            finally:
                duration_ms = (time.time_ns() - span.start_ns) / 1_000_000.0
                span.set_attribute("request_duration_ms", duration_ms)


_default_session: Optional[TracedSession] = None


def get_traced_session() -> TracedSession:
    global _default_session
    if _default_session is None:
        _default_session = TracedSession()
    return _default_session


def _record_outbound_json_preview(response: requests.Response, span: TraceSpan) -> None:
    """If response is JSON, attach small part of it to span"""
    try:
        content_type = str(response.headers.get('Content-Type', '')).lower()
        if 'application/json' not in content_type:
            return
        text = response.text  # requests will decode using inferred/declared encoding
        span.set_attribute('http.response.body.size', len(text))
        span.set_attribute('http.response.json', text[:_PREVIEW_MAX_CHARS])
    except Exception:
        logger.exception("Could not extract JSON response body")



import logging
from typing import Any, Dict, List, Optional, Union

import pyotp
import requests


from cmapi_server.constants import (
    CMAPI_CONF_PATH, CURRENT_NODE_CMAPI_URL, SECRET_KEY, _version
)
from cmapi_server.exceptions import CMAPIBasicError
from cmapi_server.helpers import get_config_parser, get_current_key
from tracing.traced_session import get_traced_session


class BaseClient:
    """Base class for API clients.

    This class is not intended to be used directly, but rather as a
    base class for other API clients. It provides a common interface
    for making requests to the API and handling responses.
    WARNING: This class only handles the API requests, it does not
             handle the transaction management. So it should be started
             at level above using TransactionManager (decorator or context
             manager).
    """
    def __init__(
        self, base_url: str = CURRENT_NODE_CMAPI_URL, request_timeout: Optional[float] = None
    ):
        """Initialize the BaseClient with the base URL.

        :param base_url: The base URL for the API endpoints,
                         defaults to CURRENT_NODE_CMAPI_URL
        """
        self.base_url = base_url
        self.request_timeout = request_timeout
        self.cmd_class = None

    def _request(
        self, method: str, endpoint: str, data: Optional[Dict[str, Any]] = None,
        throw_real_exp: bool = False
    ) -> Union[Dict[str, Any], List[Any]]:
        """Make a request to the API.

        :param method: The HTTP method to use.
        :param endpoint: The API endpoint to call.
        :param data: The data to send with the request.
        :return: The response from the API.
        """

        url = f'{self.base_url}/cmapi/{_version}/{self.cmd_class}/{endpoint}'
        cmapi_cfg_parser = get_config_parser(CMAPI_CONF_PATH)
        key = get_current_key(cmapi_cfg_parser)
        headers = {'x-api-key': key}
        if method in ['PUT', 'POST', 'DELETE']:
            headers['Content-Type'] = 'application/json'
            data = {'in_transaction': True, **(data or {})}
        try:
            response = get_traced_session().request(
                method, url, headers=headers,
                params=data if method == 'GET' else None,
                json=data if method in ('PUT', 'POST', 'DELETE') else None,
                timeout=self.request_timeout, verify=False
            )
            response.raise_for_status()
            return response.json()
        except requests.exceptions.ConnectionError as exc:
            message = (
                f'API client could not connect to {url}. '
                'Is cmapi server running and reachable?'
            )
            logging.error(message)
            if throw_real_exp:
                raise exc
            raise CMAPIBasicError(message)
        except requests.HTTPError as exc:
            resp = exc.response
            request_url = exc.request.url if exc.request else url
            status_code = resp.status_code if resp is not None else 'N/A'
            error_msg = str(exc)
            if status_code == 422:
                # in this case we think cmapi server returned some value but
                # had error during running endpoint handler code
                try:
                    resp_json = resp.json()
                    error_msg = resp_json.get('error', resp_json)
                except requests.exceptions.JSONDecodeError:
                    error_msg = resp.text
            message = (
                f'API client got an HTTPError exception in request to {request_url} with code '
                f'{status_code} and error: {error_msg}'
            )
            logging.error(message)
            if throw_real_exp:
                raise exc
            raise CMAPIBasicError(message)
        except requests.exceptions.Timeout:
            message = f'Request to {url} timed out after {self.request_timeout}'
            logging.error(message)
            if throw_real_exp:
                raise exc
            raise CMAPIBasicError(message)
        except requests.exceptions.RequestException as exc:
            resp = exc.response
            request_url = exc.request.url if exc.request else url
            status_code = resp.status_code if exc.response is not None else 'N/A'
            message = (
                'API client got an undefined error in request to '
                f'{request_url} with code {status_code!r} and '
                f'error: {str(exc)}'
            )
            logging.error(message)
            if throw_real_exp:
                raise exc
            raise CMAPIBasicError(message)


class ClusterControllerClient(BaseClient):
    """Client for the ClusterController API.
    This class provides methods for interacting with the cluster
    management API, including starting and stopping the cluster,
    adding and removing nodes, and getting the cluster status.
    """
    def __init__(
        self, base_url: str = CURRENT_NODE_CMAPI_URL, request_timeout: Optional[float] = None
    ):
        """Initialize the BaseClient with the base URL.

        :param base_url: The base URL for the API endpoints,
                         defaults to CURRENT_NODE_CMAPI_URL
        :type base_url: str, optional
        :param request_timeout: request timeout, defaults to None
        :type request_timeout: Optional[float], optional
        """
        super().__init__(base_url, request_timeout)
        self.cmd_class = 'cluster'

    def start_cluster(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Start the cluster.

        :return: The response from the API.
        """
        return self._request('PUT', 'start', extra)

    def shutdown_cluster(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Shutdown the cluster.

        :return: The response from the API.
        """
        return self._request('PUT', 'shutdown', extra)

    def set_mode(
        self, mode: str, extra: Optional[Dict[str, Any]] = None
        ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Set the cluster mode.

        :param mode: The mode to set.
        :return: The response from the API.
        """
        return self._request('PUT', 'mode-set', {'mode': mode, **(extra or {})})

    def add_node(
        self, node_info: Dict[str, Any], extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Add a node to the cluster.

        :param node_info: Information about a node to add.
        :return: The response from the API.
        """
        #TODO: fix interface as in remove_node used or think about universal
        return self._request('PUT', 'node', {**node_info, **(extra or {})})

    def remove_node(
        self, node: str, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Remove a node from the cluster.

        :param node: node IP, name or FQDN.
        :return: The response from the API.
        """
        return self._request('DELETE', 'node', {'node': node, **(extra or {})})

    def get_status(self) -> Union[Dict[str, Any], Dict[str, str]]:
        """Get the status of the cluster.

        :return: The response from the API.
        """
        return self._request('GET', 'status')

    def get_health(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """
        Get the health of the cluster.

        :return: The response from the API.
        """
        return self._request('GET', 'health', extra)

    def set_api_key(self, api_key: str) -> Union[Dict[str, Any], Dict[str, str]]:
        """Set the API key for the cluster.

        :param api_key: The API key to set.
        :return: The response from the API.
        """
        totp = pyotp.TOTP(SECRET_KEY)
        payload = {
            'api_key': api_key,
            'verification_key': totp.now()
        }
        return self._request('PUT', 'apikey-set', payload)

    def set_log_level(self, log_level: str) -> Union[Dict[str, Any], Dict[str, str]]:
        """Set the log level for the cluster.

        :param log_level: The log level to set.
        :return: The response from the API.
        """
        return self._request('PUT', 'log-level', {'log_level': log_level})

    def load_s3data(self, s3data_info: Dict[str, Any]) -> Union[Dict[str, Any], Dict[str, str]]:
        """Load S3 data into the cluster.

        :param s3data_info: Information about the S3 data to load.
        :return: The response from the API.
        """
        return self._request('PUT', 'load_s3data', s3data_info)

    def get_versions(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Get packages versions.
        :return: The response from the API.
        """
        return self._request('GET', 'versions', extra)

    def start_mariadb(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Start MariaDB server service on each node in cluster.

        :return: The response from the API.
        """
        return self._request('PUT', 'start-mariadb', extra)

    def stop_mariadb(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Stop MariaDB server service on each node in cluster.

        :return: The response from the API.
        """
        return self._request('PUT', 'stop-mariadb', extra)

    def install_repo(
        self, token: str, mariadb_version: str, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Install ES repository on each node in cluster.

        :return: The response from the API.
        """
        data = {
            'token': token,
            'mariadb_version': mariadb_version
        }
        return self._request('PUT', 'install-repo', {**data, **(extra or {})})

    def preupgrade_backup(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Backup DBRM and configs on each node in cluster.

        :return: The response from the API.
        """
        return self._request('PUT', 'preupgrade-backup', extra)

    def upgrade_mdb_mcs(
        self, mariadb_version: str, columnstore_version: str,
        extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Upgrade MariaDB and Columnstore on each node in cluster.

        :return: The response from the API.
        """
        data = {
            'mariadb_version': mariadb_version,
            'columnstore_version': columnstore_version
        }
        return self._request('PUT', 'upgrade-mdb-mcs', {**data, **(extra or {})})

    def upgrade_cmapi(
        self, version: str, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Starts cmapi_updater.service on each node and waits for new cmapi.

        :return: The response from the API.
        """
        return self._request(
            'PUT', 'upgrade-cmapi', {'version': version, **(extra or {})}
        )

    def check_shared_storage(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Check if shared storage working.

        :return: The response from the API.
        """
        return self._request('PUT', 'check-shared-storage', extra)


class NodeControllerClient(BaseClient):
    """Client for the NodeController API.

    This class provides methods for interacting with a node management
    API.
    """
    def __init__(
        self, base_url: str = CURRENT_NODE_CMAPI_URL, request_timeout: Optional[float] = None
    ):
        """Initialize the NodeControllerClient with the base URL.

        :param base_url: The base URL for the API endpoints,
                         defaults to CURRENT_NODE_CMAPI_URL
        :type base_url: str, optional
        :param request_timeout: request timeout, defaults to None
        :type request_timeout: Optional[float], optional
        """
        super().__init__(base_url, request_timeout)
        self.cmd_class = 'node'

    def get_versions(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Get packages versions installed on a node.

        :return: The response from the API.
        """
        return self._request('GET', 'versions', extra)

    def get_latest_mdb_version(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Get latest tested MDB version from repo.

        :return: The response from the API.
        """
        return self._request('GET', 'latest-mdb-version', extra)

    def validate_mdb_version(
        self, token: str, mariadb_version: str, extra: Optional[Dict[str, Any]] = None, **kwargs
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Verify MariaDB ES version.

        :param token: valid ES token
        :type token: str
        :param mariadb_version: MariaDB version to verify
        :type mariadb_version: str
        :return: The response from the API
        :rtype: Union[Dict[str, Any], Dict[str, str]]
        """
        data = {
            'token': token,
            'mariadb_version': mariadb_version
        }
        return self._request('GET', 'validate-mdb-version', {**data, **(extra or {})}, **kwargs)

    def validate_es_token(
        self, token: str, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Verify MariaDB ES token is correct.

        :param token: ES token to verify
        :type token: str
        :return: The response from the API.
        """
        return self._request(
            'GET', 'validate-es-token', {'token': token, **(extra or {})}
        )

    def start_mariadb(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Start MariaDB-server service on a node.

        :return: The response from the API.
        """
        return self._request('PUT', 'start-mariadb', extra)

    def stop_mariadb(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Stop MariaDB-server service on a node.

        :return: The response from the API.
        """
        return self._request('PUT', 'stop-mariadb', extra)

    def is_process_running(
        self, process_name: str, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Check if a process is running on a node.

        :return: The response from the API.
        """
        return self._request(
            'GET', 'is-process-running', {'process_name': process_name, **(extra or {})}
        )

    def repo_pkg_versions(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Get available packages versions from the repo on a node.

        :return: The response from the API.
        """
        return self._request('GET', 'repo-pkg-versions', extra)

    def install_repo(
        self, token: str, mariadb_version: str, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Install the repository on a node.

        :param token: valid ES token
        :type token: str
        :param mariadb_version: MariaDB version to verify
        :type mariadb_version: str
        :return: The response from the API.
        """
        data = {
            'token': token,
            'mariadb_version': mariadb_version
        }
        return self._request('PUT', 'install-repo', {**data, **(extra or {})})


    def preupgrade_backup(
        self, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Backup DBRM and configs on a node.

        :return: The response from the API.
        """
        return self._request('PUT', 'preupgrade-backup', extra)

    def upgrade_mdb_mcs(
        self, mariadb_version: str, columnstore_version: str,
        extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Upgrade MariaDB and Columnstore on a node.

        :return: The response from the API.
        """
        data = {
            'mariadb_version': mariadb_version,
            'columnstore_version': columnstore_version
        }
        return self._request('PUT', 'upgrade-mdb-mcs', {**data, **(extra or {})})

    def kick_cmapi_upgrade(
        self, version: str, extra: Optional[Dict[str, Any]] = None
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Starting cmapi_updater.service on a node.

        :return: The response from the API.
        """
        return self._request(
            'PUT', 'kick-cmapi-upgrade', {'version': version, **(extra or {})}
        )

    def check_shared_file(
        self, file_path: str, check_sum: str
    ) -> Union[Dict[str, Any], Dict[str, str]]:
        """Get packages versions installed on a node.

        :param file_path: file path to check
        :type file_path: str
        :param check_sum: expected MD5 file checksum
        :type check_sum: str
        :return: The response from the API.
        """
        data = {
            'file_path': file_path,
            'check_sum': check_sum,
        }
        return self._request('GET', 'check-shared-file', data)



class AppControllerClient(BaseClient):
    """Client for the AppController API.

    This class provides methods for interacting with a cmapi special management
    API.
    """
    def __init__(
        self, base_url: str = CURRENT_NODE_CMAPI_URL, request_timeout: Optional[float] = None
    ):
        """Initialize the NodeControllerClient with the base URL.

        :param base_url: The base URL for the API endpoints,
                         defaults to CURRENT_NODE_CMAPI_URL
        :type base_url: str, optional
        :param request_timeout: request timeout, defaults to None
        :type request_timeout: Optional[float], optional
        """
        super().__init__(base_url, request_timeout)
        self.cmd_url = f'{self.base_url}/cmapi/'

    def get_ready(self) -> Union[Dict[str, Any], Dict[str, str]]:
        """Get CMAPI ready or not.

        :return: The response from the API.
        """
        return self._request('GET', 'ready', None, throw_real_exp=True)



import logging
import os
import secrets
import sys
import time
from datetime import datetime, timedelta
from typing import Optional
import ast
from collections import Counter

import requests
import typer
from typing_extensions import Annotated
from rich.console import Console
from rich.progress import (
    BarColumn, Progress, SpinnerColumn, TimeElapsedColumn,
)
from rich.table import Table


from cmapi_server.constants import (
    CMAPI_CONF_PATH,
    CMAPI_PORT,
    MCS_DATA_PATH,
    MCS_SECRETS_FILENAME,
    REQUEST_TIMEOUT,
    TRANSACTION_TIMEOUT,
)
from cmapi_server.controllers.api_clients import (
    AppControllerClient, ClusterControllerClient, NodeControllerClient
)
from cmapi_server.exceptions import CEJError, CMAPIBasicError
from cmapi_server.handlers.cej import CEJPasswordHandler
from cmapi_server.helpers import get_active_nodes, get_config_parser
from cmapi_server.managers.transaction import TransactionManager
from cmapi_server.managers.upgrade.utils import ComparableVersion
from cmapi_server.process_dispatchers.base import BaseDispatcher
from mcs_cluster_tool.constants import MCS_COLUMNSTORE_REVIEW_SH, INSTALL_ES_LOG_FILEPATH
from mcs_cluster_tool.decorators import handle_output
from mcs_cluster_tool.helpers import cook_sh_arg



logger = logging.getLogger('mcs_cli')
# pylint: disable=unused-argument, too-many-arguments, too-many-locals
# pylint: disable=invalid-name, line-too-long


@handle_output
def cskeys(
    user: Annotated[
        str,
        typer.Option(
            '-u', '--user',
            help='Designate the owner of the generated file.',
        )
    ] = 'mysql',
    directory: Annotated[
        str,
        typer.Argument(
            help='The directory where to store the file in.',
        )
    ] = MCS_DATA_PATH
):
    """
    This utility generates a random AES encryption key and init vector
    and writes them to disk. The data is written to the file '.secrets',
    in the specified directory. The key and init vector are used by
    the utility 'cspasswd' to encrypt passwords used in Columnstore
    configuration files, as well as by Columnstore itself to decrypt the
    passwords.

    WARNING: Re-creating the file invalidates all existing encrypted
    passwords in the configuration files.
    """
    filepath = os.path.join(directory, MCS_SECRETS_FILENAME)
    if CEJPasswordHandler().secretsfile_exists(directory=directory):
        typer.echo(
            (
                f'Secrets file "{filepath}" already exists. '
                'Delete it before generating a new encryption key.'
            ),
            color='red',
        )
        raise typer.Exit(code=1)
    elif not os.path.exists(os.path.dirname(filepath)):
        typer.echo(
            f'Directory "{directory}" does not exist.',
            color='red'
        )
        raise typer.Exit(code=1)

    new_secrets_data = CEJPasswordHandler().generate_secrets_data()
    try:
        CEJPasswordHandler().save_secrets(
            new_secrets_data, owner=user, directory=directory
        )
        typer.echo(f'Permissions of "{filepath}" set to owner:read.')
        typer.echo(f'Ownership of "{filepath}" given to {user}.')
    except CEJError as cej_error:
        typer.echo(cej_error.message, color='red')
        raise typer.Exit(code=2)
    raise typer.Exit(code=0)


@handle_output
def cspasswd(
    password: Annotated[
        str,
        typer.Option(
            help='Password to encrypt/decrypt',
            prompt=True, confirmation_prompt=True, hide_input=True
        )
    ],
    decrypt: Annotated[
        bool,
        typer.Option(
            '--decrypt',
            help='Decrypt an encrypted password instead.',
        )
    ] = False
):
    """
    Encrypt a Columnstore plaintext password using the encryption key in
    the key file.
    """
    if decrypt:
        try:
            decrypted_password = CEJPasswordHandler().decrypt_password(
                password
            )
        except CEJError as cej_error:
            typer.echo(cej_error.message, color='red')
            raise typer.Exit(code=1)
        typer.echo(f'Decoded password: {decrypted_password}', color='green')
    else:
        try:
            encoded_password = CEJPasswordHandler().encrypt_password(password)
        except CEJError as cej_error:
            typer.echo(cej_error.message, color='red')
            raise typer.Exit(code=1)
        typer.echo(f'Encoded password: {encoded_password}', color='green')
    raise typer.Exit(code=0)


@handle_output
def bootstrap_single_node(
    key: Annotated[
        str,
        typer.Option(
            '--api-key',
            help='API key to set.',
        )
    ] = ''
):
    """Bootstrap a single node (localhost) Columnstore instance."""
    node = 'localhost'
    client = ClusterControllerClient(request_timeout=REQUEST_TIMEOUT)
    if not key:
        # Generate API key if not provided
        key = secrets.token_urlsafe(32)  #pylint: disable=no-member
    # handle_output decorator will catch, show and log errors
    api_key_set_resp = client.set_api_key(key)
    # if operation takes minutes, then it is better to raise by timeout
    with TransactionManager(
        timeout=TRANSACTION_TIMEOUT, handle_signals=True,
        extra_nodes=[node]
    ):
        add_node_resp = client.add_node({'node': node})

    result = {
        'timestamp': str(datetime.now()),
        'set_api_key_resp': api_key_set_resp,
        'add_node_resp': add_node_resp,
    }
    return result


@handle_output
def review(
    _version: Annotated[
        Optional[bool],
        typer.Option(
            '--version',
            help='Only show the header with version information.',
            show_default=False
        )
    ] = None,
    _logs: Annotated[
        Optional[bool],
        typer.Option(
            '--logs',
            help=(
                'Create a compressed archive of logs for MariaDB Support '
                'Ticket'
            ),
            show_default=False
        )
    ] = None,
    _path: Annotated[
        Optional[str],
        typer.Option(
            '--path',
            help=(
                'Define the path for where to save files/tarballs and outputs '
                'of this script.'
            ),
            show_default=False
        )
    ] = None,
    _backupdbrm: Annotated[
        Optional[bool],
        typer.Option(
            '--backupdbrm',
            help=(
                'Takes a compressed backup of extent map files in dbrm '
                'directory.'
            ),
            show_default=False
        )
    ] = None,
    _testschema: Annotated[
        Optional[bool],
        typer.Option(
            '--testschema',
            help=(
                'Creates a test schema, tables, imports, queries, drops '
                'schema.'
            ),
            show_default=False
        )
    ] = None,
    _testschemakeep: Annotated[
        Optional[bool],
        typer.Option(
            '--testschemakeep',
            help=(
                'Creates a test schema, tables, imports, queries, does not '
                'drop.'
            ),
            show_default=False
        )
    ] = None,
    _ldlischema: Annotated[
        Optional[bool],
        typer.Option(
            '--ldlischema',
            help=(
                'Using ldli, creates test schema, tables, imports, queries, '
                'drops schema.'
            ),
            show_default=False
        )
    ] = None,
    _ldlischemakeep: Annotated[
        Optional[bool],
        typer.Option(
            '--ldlischemakeep',
            help=(
                'Using ldli, creates test schema, tables, imports, queries, '
                'does not drop.'
            ),
            show_default=False
        )
    ] = None,
    _emptydirs: Annotated[
        Optional[bool],
        typer.Option(
            '--emptydirs',
            help='Searches /var/lib/columnstore for empty directories.',
            show_default=False
        )
    ] = None,
    _notmysqldirs: Annotated[
        Optional[bool],
        typer.Option(
            '--notmysqldirs',
            help=(
                'Searches /var/lib/columnstore for directories not owned by '
                'mysql.'
            ),
            show_default=False
        )
    ] = None,
    _emcheck: Annotated[
        Optional[bool],
        typer.Option(
            '--emcheck',
            help='Checks the extent map for orphaned and missing files.',
            show_default=False
        )
    ] = None,
    _s3check: Annotated[
        Optional[bool],
        typer.Option(
            '--s3check',
            help='Checks the extent map against S3 storage.',
            show_default=False
        )
    ] = None,
    _pscs: Annotated[
        Optional[bool],
        typer.Option(
            '--pscs',
            help=(
                'Adds the pscs command. pscs lists running columnstore '
                'processes.'
            ),
            show_default=False
        )
    ] = None,
    _schemasync: Annotated[
        Optional[bool],
        typer.Option(
            '--schemasync',
            help='Fix out-of-sync columnstore tables (CAL0009).',
            show_default=False
        )
    ] = None,
    _tmpdir: Annotated[
        Optional[bool],
        typer.Option(
            '--tmpdir',
            help=(
                'Ensure owner of temporary dir after reboot (MCOL-4866 & '
                'MCOL-5242).'
            ),
            show_default=False
        )
    ] = None,
    _checkports: Annotated[
        Optional[bool],
        typer.Option(
            '--checkports',
            help='Checks if ports needed by Columnstore are opened.',
            show_default=False
        )
    ] = None,
    _eustack: Annotated[
        Optional[bool],
        typer.Option(
            '--eustack',
            help='Dumps the stack of Columnstore processes.',
            show_default=False
        )
    ] = None,
    _clearrollback: Annotated[
        Optional[bool],
        typer.Option(
            '--clearrollback',
            help='Clear any rollback fragments from dbrm files.',
            show_default=False
        )
    ] = None,
    _killcolumnstore: Annotated[
        Optional[bool],
        typer.Option(
            '--killcolumnstore',
            help=(
                'Stop columnstore processes gracefully, then kill remaining '
                'processes.'
            ),
            show_default=False
        )
    ] = None,
    _color: Annotated[
        Optional[str],
        typer.Option(
            '--color',
            help=(
                'print headers in color. Options: [none,red,blue,green,yellow,'
                'magenta,cyan, none] prefix color with l for light.'
            ),
            show_default=False
        )
    ] = None,
):
    """
    This script performs various maintenance and diagnostic tasks for
    MariaDB ColumnStore, including log archiving, extent map backups,
    schema and table testing, directory and ownership checks, extent map
    validation, S3 storage comparison, process management, table
    synchronization, port availability checks, stack dumps, cleanup of
    rollback fragments, and graceful process termination.

    If database is up, this script will connect as root@localhost via socket.
    """

    arguments = []
    for arg_name, value in locals().items():
        sh_arg = cook_sh_arg(arg_name, value, separator='=')
        if sh_arg is None:
            continue
        # columnstore_review.sh accepts only --arg=value format
        arguments.append(sh_arg)
    cmd = f'{MCS_COLUMNSTORE_REVIEW_SH} {" ".join(arguments)}'
    success, _ = BaseDispatcher.exec_command(cmd, stdout=sys.stdout)
    if not success:
        raise typer.Exit(code=1)
    raise typer.Exit(code=0)


# Sentry subcommand app
sentry_app = typer.Typer(help='Manage Sentry DSN configuration for error tracking.')


@sentry_app.command()
@handle_output
def show():
    """Show current Sentry DSN configuration."""
    try:
        # Read existing config
        cfg_parser = get_config_parser(CMAPI_CONF_PATH)

        if not cfg_parser.has_section('Sentry'):
            typer.echo('Sentry is disabled (no configuration found).', color='yellow')
            raise typer.Exit(code=0)

        dsn = cfg_parser.get('Sentry', 'dsn', fallback='').strip().strip("'\"")
        environment = cfg_parser.get('Sentry', 'environment', fallback='development').strip().strip("'\"")

        if not dsn:
            typer.echo('Sentry is disabled (DSN is empty).', color='yellow')
        else:
            typer.echo('Sentry is enabled:', color='green')
            typer.echo(f'  DSN: {dsn}')
            typer.echo(f'  Environment: {environment}')

    except Exception as e:
        typer.echo(f'Error reading configuration: {str(e)}', color='red')
        raise typer.Exit(code=1)

    raise typer.Exit(code=0)


@sentry_app.command()
@handle_output
def enable(
    dsn: Annotated[
        str,
        typer.Argument(
            help='Sentry DSN URL to enable for error tracking.',
        )
    ],
    environment: Annotated[
        str,
        typer.Option(
            '--environment', '-e',
            help='Sentry environment name (default: development).',
        )
    ] = 'development'
):
    """Enable Sentry error tracking with the provided DSN."""
    if not dsn:
        typer.echo('DSN cannot be empty.', color='red')
        raise typer.Exit(code=1)

    try:
        # Read existing config
        cfg_parser = get_config_parser(CMAPI_CONF_PATH)

        # Add or update Sentry section
        if not cfg_parser.has_section('Sentry'):
            cfg_parser.add_section('Sentry')

        cfg_parser.set('Sentry', 'dsn', f"'{dsn}'")
        cfg_parser.set('Sentry', 'environment', f"'{environment}'")

        # Write config back to file
        with open(CMAPI_CONF_PATH, 'w') as config_file:
            cfg_parser.write(config_file)

        typer.echo('Sentry error tracking enabled successfully.', color='green')
        typer.echo(f'DSN: {dsn}', color='green')
        typer.echo(f'Environment: {environment}', color='green')
        typer.echo('Note: Restart cmapi service for changes to take effect.', color='yellow')

    except Exception as e:
        typer.echo(f'Error updating configuration: {str(e)}', color='red')
        raise typer.Exit(code=1)

    raise typer.Exit(code=0)


@sentry_app.command()
@handle_output
def disable():
    """Disable Sentry error tracking by removing the configuration."""
    try:
        # Read existing config
        cfg_parser = get_config_parser(CMAPI_CONF_PATH)

        if not cfg_parser.has_section('Sentry'):
            typer.echo('Sentry is already disabled (no configuration found).', color='yellow')
            raise typer.Exit(code=0)

        # Remove the entire Sentry section
        cfg_parser.remove_section('Sentry')

        # Write config back to file
        with open(CMAPI_CONF_PATH, 'w') as config_file:
            cfg_parser.write(config_file)

        typer.echo('Sentry error tracking disabled successfully.', color='green')
        typer.echo('Note: Restart cmapi service for changes to take effect.', color='yellow')

    except Exception as e:
        typer.echo(f'Error updating configuration: {str(e)}', color='red')
        raise typer.Exit(code=1)

    raise typer.Exit(code=0)


@handle_output
def healthcheck():
    """Check the health of the MCS cluster."""
    with TransactionManager(
        timeout=timedelta(minutes=5).total_seconds(), handle_signals=True,
    ):
        client = ClusterControllerClient(request_timeout=REQUEST_TIMEOUT)
        result = client.get_health({'in_transaction': True})
        # TODO: just a placeholder for now
        #       need to implement result in a table format with color
        typer.echo(
            'Cluster health check completed successfully.',
            color='green'
        )
    raise typer.Exit(code=0)


@handle_output
def install_es(
    token: Annotated[
        str,
        typer.Option(
            '--token',
            help='ES API Token to use for the upgrade.',
            show_default=False
        )
    ],
    target_version: Annotated[
        str,
        typer.Option(
            '-v', '--version',
            help='ES version to upgdate.',
            show_default=False
        )
    ] = 'latest',
    ignore_mismatch: Annotated[
        bool,
        typer.Option(
            '--ignore-mismatch',
            help='Proceed even if nodes report different installed package versions (use majority as baseline).',
            show_default=False
        )
    ] = False,
):
    """
    [Beta]
    Install the specified MDB ES version.
    If the version is 'latest', it will upgrade to the latest tested version
    available.
    """
    new_handler = logging.FileHandler(INSTALL_ES_LOG_FILEPATH, mode='w')
    new_handler.setLevel(logging.DEBUG)
    new_handler.setFormatter(logging.getLogger('mcs_cli').handlers[0].formatter)
    for logger_name in ("", "mcs_cli"):
        current_logger = logging.getLogger(logger_name)
        current_logger.addHandler(new_handler)
    console = Console()
    console.clear()
    console.rule('[bold green][Beta] MariaDB ES Installer')

    console.print('This utility is now in Beta.', style='yellow underline')
    console.print(
        (
            'Downgrades are supported up to MariaDB 10.6.9-5 and Columnstore 22.08.4.'
            'Make sure you have a backup of your data before proceeding. '
            'If you encounter any issues, please report them to MariaDB Support.'
        ),
        style='underline'
    )

    # Collect output (tables/messages) to render AFTER the progress bar finishes
    post_output: list = []  # items can be strings with rich markup or Rich renderables
    exit_code: int = 0
    def post_print(msg: str, color: Optional[str] = None):
        if color:
            post_output.append(f'[{color}]{msg}[/{color}]')
        else:
            post_output.append(msg)

    active_nodes = get_active_nodes()

    node_api_client = NodeControllerClient()
    cluster_api_client = ClusterControllerClient()
    app_api_client = AppControllerClient()


    node_api_client.validate_es_token(token)
    if target_version == 'latest':
        response = node_api_client.get_latest_mdb_version()
        target_version = response['latest_mdb_version']
    else:
        try:
            node_api_client.validate_mdb_version(token, target_version, throw_real_exp=True)
        except requests.exceptions.HTTPError as exc:
            resp = exc.response
            error_msg = str(exc)
            if resp.status_code == 422:
                try:
                    resp_json = resp.json()
                    error_msg = resp_json.get('error', resp_json)
                except requests.exceptions.JSONDecodeError:
                    error_msg = resp.text
            console.print('ERROR:', style='red')
            console.print(error_msg, style='underline')
            console.rule()
            raise typer.Exit(code=1)

    # Retrieve current versions; if nodes are mismatched, prettify the server error.
    # If --ignore-mismatch is passed we will continue, choosing the majority version
    # of each package as the baseline "current" version.
    try:
        versions = cluster_api_client.get_versions()
    except CMAPIBasicError as exc:  # custom API client error
        msg = exc.message
        mismatch_marker = 'Packages versions:'
        if mismatch_marker in msg:
            try:
                dict_part = msg.split(mismatch_marker, 1)[1].strip()
                packages_versions = ast.literal_eval(dict_part)
            except Exception:  # pragma: no cover - defensive
                # Could not parse, fall back to original behavior
                console.print(f"[red]{msg}[/red]")
                raise typer.Exit(code=1)

            console.print('Detected package version mismatch across nodes:', style='yellow')
            mismatch_table = Table('Node', 'Server', 'Columnstore', 'CMAPI')

            server_vals = [v.get('server_version') for v in packages_versions.values()]
            cs_vals = [v.get('columnstore_version') for v in packages_versions.values()]
            cmapi_vals = [v.get('cmapi_version') for v in packages_versions.values()]
            server_common = Counter(server_vals).most_common(1)[0][0] if server_vals else None
            cs_common = Counter(cs_vals).most_common(1)[0][0] if cs_vals else None
            cmapi_common = Counter(cmapi_vals).most_common(1)[0][0] if cmapi_vals else None

            def style(val, common):
                if val is None:
                    return '[red]-[/red]'
                return f'[green]{val}[/green]' if val == common else f'[red]{val}[/red]'

            for node, vers in sorted(packages_versions.items()):
                mismatch_table.add_row(
                    node,
                    style(vers.get('server_version'), server_common),
                    style(vers.get('columnstore_version'), cs_common),
                    style(vers.get('cmapi_version'), cmapi_common),
                )
            # Print after progress unless we're going to exit early
            if not ignore_mismatch:
                # No progress has started yet; render now and exit
                console.print(mismatch_table)
                console.print('[yellow]All nodes must have identical package versions before running install-es. '
                               'Please align versions (upgrade/downgrade individual nodes) and retry, '
                               'or rerun with --ignore-mismatch to force.[/yellow]')
                raise typer.Exit(code=1)

            console.print(mismatch_table)
            # Forced continuation path
            console.print(
                (
                    'Proceeding despite mismatch ( --ignore-mismatch ). Using majority versions '
                    'as baseline.'
                ),
                style='yellow'
            )
            versions = {
                'server_version': server_common or server_vals[0],
                'columnstore_version': cs_common or cs_vals[0],
                'cmapi_version': cmapi_common or cmapi_vals[0],
            }
        else:
            # Not a mismatch we recognize; rethrow for decorator to handle
            raise
    mdb_curr_ver = versions['server_version']
    mcs_curr_ver = versions['columnstore_version']
    cmapi_curr_ver = versions['cmapi_version']
    mdb_curr_ver_comp = ComparableVersion(mdb_curr_ver)
    mdb_target_ver_comp = ComparableVersion(target_version)

    console.print('Currently installed vesions:', style='green')
    table = Table('ES version', 'Columnstore version', 'CMAPI version')
    table.add_row(mdb_curr_ver, mcs_curr_ver, cmapi_curr_ver)
    console.print(table)
    is_downgrade = False
    if mdb_curr_ver_comp == mdb_target_ver_comp:
        console.print('[green]The target MariaDB ES version is already installed.[/green]')
        raise typer.Exit(code=0)
    elif mdb_curr_ver_comp > mdb_target_ver_comp:
        downgrade = typer.confirm(
            'Target version is older than currently installed. '
            'Are you sure you really want to downgrade?\n'
            'WARNING: Could cause data loss and/or broken cluster.',
            prompt_suffix=' '
        )
        if not downgrade:
            raise typer.Exit(code=1)
        is_downgrade = True
    elif mdb_curr_ver_comp < mdb_target_ver_comp:
        upgrade = typer.confirm(
            f'Are you sure you really want to upgrade to {target_version}?',
            prompt_suffix=' '
        )
        if not upgrade:
            raise typer.Exit(code=1)

    if not active_nodes:
        post_print('No active nodes found, using localhost only.', 'yellow')
        active_nodes.append('localhost')

    with Progress(
        SpinnerColumn(),
        '[progress.description]{task.description}',
        BarColumn(),
        TimeElapsedColumn(),
        console=console,
    ) as progress:
        step1_stop_cluster = progress.add_task('Stopping MCS cluster...', total=None)
        with TransactionManager(
            timeout=timedelta(days=1).total_seconds(), handle_signals=True
        ):
            cluster_stop_resp = cluster_api_client.shutdown_cluster(
                {'in_transaction': True}
            )
        progress.update(
            step1_stop_cluster, description='[green]MCS Cluster stopped ✓', total=100,
            completed=True
        )
        progress.stop_task(step1_stop_cluster)

        step2_stop_mariadb = progress.add_task('Stopping MariaDB server...', total=None)
        # TODO: put MaxScale into maintainance mode
        mariadb_stop_resp = cluster_api_client.stop_mariadb(
            {'in_transaction': True}
        )
        progress.update(
            step2_stop_mariadb, description='[green]MariaDB server stopped ✓', total=100,
            completed=True
        )
        progress.stop_task(step2_stop_mariadb)

        step3_install_es_repo = progress.add_task(
            'Installing MariaDB ES repository...', total=None
        )
        inst_repo_response = cluster_api_client.install_repo(
            token=token, mariadb_version=target_version
        )
        progress.update(
            step3_install_es_repo, description='[green]Repository installed ✓', total=100,
            completed=True
        )
        progress.stop_task(step3_install_es_repo)

        if target_version == 'latest':
            # PackageManager accepts latest versions so no need to get numeric
            mdb_target_ver = mcs_target_ver = cmapi_target_ver = 'latest'
        else:
            step3_5_get_available_versions = progress.add_task(
                'Getting available versions of packages...', total=None
            )
            available_versions_resp = node_api_client.repo_pkg_versions()
            mdb_target_ver = available_versions_resp['server_version']
            mcs_target_ver = available_versions_resp['columnstore_version']
            cmapi_target_ver = available_versions_resp['cmapi_version']
            progress.update(
                step3_5_get_available_versions,
                description=(
                    f'[green]Available versions: ES {mdb_target_ver}, '
                    f'Columnstore {mcs_target_ver}, CMAPI {cmapi_target_ver} ✓'
                ),
                total=100, completed=True
            )
            progress.stop_task(step3_5_get_available_versions)

        step4_preupgrade_backup = progress.add_task(
            'Starting pre-upgrade backup DBRM and configs on each node...', total=None
        )
        backup_response = cluster_api_client.preupgrade_backup()
        progress.update(
            step4_preupgrade_backup, description='[green]PreUpgrade Backup completed ✓',
            total=100, completed=True
        )
        progress.stop_task(step4_preupgrade_backup)

        step5_upgrade_mdb_mcs = progress.add_task(
            'Upgrading MariaDB and Columnstore on each node...', total=None
        )
        mdb_mcs_upgrade_response = cluster_api_client.upgrade_mdb_mcs(
            mariadb_version=mdb_target_ver, columnstore_version=mcs_target_ver
        )
        progress.update(
            step5_upgrade_mdb_mcs,
            description=f'[green]Upgraded to MariaDB {mdb_target_ver} and Columnstore {mcs_target_ver} ✓',
            total=100, completed=True
        )
        progress.stop_task(step5_upgrade_mdb_mcs)

        step6_install_cmapi = progress.add_task('Upgrading CMAPI on each node...', total=None)
        try:
            cmapi_upgrade_response = cluster_api_client.upgrade_cmapi(version=cmapi_target_ver)
            # cmapi_updater service has 5 s timeout to give CMAPI time to handle response,
            # we need to wait when API become unreachable after CMAPI stop.
            time.sleep(6)
        except requests.exceptions.ConnectionError:
            # during upgrade the connection drop is expected
            pass

        # Prepare per-node readiness tracking
        progress.update(
            step6_install_cmapi, description='Waiting CMAPI to be ready on each node...',
            completed=None
        )
        start_time = datetime.now()
        timeout_seconds = 300
        # status per node: {'status': 'PENDING'|'READY'|'ERROR'|'TIMEOUT', 'details': str}
        node_states = {
            node: {'status': 'PENDING', 'details': ''} for node in active_nodes
        }

        # Build a dedicated client per node (localhost already covered)
        per_node_clients: dict[str, AppControllerClient] = {}
        for node in active_nodes:
            if node in ('localhost', '127.0.0.1'):
                per_node_clients[node] = AppControllerClient()
            else:
                per_node_clients[node] = AppControllerClient(
                    base_url=f'https://{node}:{CMAPI_PORT}'
                )

        ready_count_prev = -1
        while (datetime.now() - start_time) < timedelta(seconds=timeout_seconds):
            ready_count = 0
            for node, client_obj in per_node_clients.items():
                # Skip nodes that already finalized (READY or ERROR)
                if node_states[node]['status'] in ('READY', 'ERROR'):
                    if node_states[node]['status'] == 'READY':
                        ready_count += 1
                    continue
                try:
                    node_response = client_obj.get_ready()
                    if node_response.get('started') is True:
                        node_states[node]['status'] = 'READY'
                        node_states[node]['details'] = 'Service started'
                        ready_count += 1
                except requests.exceptions.HTTPError as err:
                    # 503 means not ready yet, anything else mark as ERROR
                    if err.response.status_code == 503:
                        node_states[node]['details'] = 'Starting...'
                    else:
                        node_states[node]['status'] = 'ERROR'
                        node_states[node]['details'] = f'HTTP {err.response.status_code}'
                except requests.exceptions.ConnectionError:
                    # still restarting
                    node_states[node]['details'] = 'Connection refused'
                except FileNotFoundError as fnf_err:  # pragma: no cover - defensive
                    # Transient race: config file not yet created; do not fail immediately
                    missing_path = str(fnf_err).split(":")[-1].strip()
                    node_states[node]['details'] = f'Config pending ({missing_path})'
                except Exception as err:  # pragma: no cover - defensive
                    node_states[node]['status'] = 'ERROR'
                    node_states[node]['details'] = f'Unexpected: {err}'

            # Update progress description only when count changes to reduce flicker
            if ready_count != ready_count_prev:
                progress.update(
                    step6_install_cmapi,
                    description=(
                        f'Waiting CMAPI to be ready on each node... '
                        f'({ready_count}/{len(active_nodes)} ready)'
                    ),
                    completed=None
                )
                ready_count_prev = ready_count

            if ready_count == len(active_nodes):
                break
            time.sleep(1)

        # Mark TIMEOUT for nodes still pending
        for node, state in node_states.items():
            if state['status'] == 'PENDING':
                state['status'] = 'TIMEOUT'
                state['details'] = f'Not ready after {timeout_seconds}s'

        # Display per-node table after progress ends
        status_table = Table('Node', 'Status', 'Details')
        failures = False
        for node, state in sorted(node_states.items()):
            status = state['status']
            details = state['details']
            color_map = {
                'READY': 'green',
                'PENDING': 'yellow',
                'TIMEOUT': 'red',
                'ERROR': 'red',
            }
            style = color_map.get(status, 'white')
            status_table.add_row(node, f'[{style}]{status}[/{style}]', details)
            if status in ('TIMEOUT', 'ERROR'):
                failures = True
        # Defer table rendering
        post_output.append(status_table)

        if failures:
            progress.update(
                step6_install_cmapi,
                description='[red]CMAPI did not start successfully on all nodes ✗',
                total=100,
                completed=True
            )
            progress.stop_task(step6_install_cmapi)
            exit_code = 1
            post_print('CMAPI did not start successfully on all nodes.', 'red')
        else:
            progress.update(
                step6_install_cmapi,
                description='[green]CMAPI is ready on all nodes ✓',
                total=100,
                completed=True
            )
            progress.stop_task(step6_install_cmapi)

        if failures:
            # skip any automatic restarts on failure
            pass
        elif is_downgrade:
            note_panel = Table('Action', 'Status')
            note_panel.add_row('Automatic restart (MariaDB, Cluster, Health)', '[yellow]SKIPPED (downgrade)')
            post_output.append(note_panel)
            post_print(
                'Downgrade detected: automatic service restarts were skipped. '
                'Please manually start MariaDB and the ColumnStore cluster, and verify health.',
                'yellow'
            )
            post_print('Suggested manual sequence:', 'yellow')
            post_print('  1) systemctl start mariadb', 'yellow')
            post_print('  2) Use mcs-cluster tool to start cluster if needed', 'yellow')
            exit_code = 0
        else:
            step7_start_mariadb = progress.add_task('Starting MariaDB server...', total=None)
            # TODO: put MaxScale from maintainance into working mode
            mariadb_start_resp = cluster_api_client.start_mariadb({'in_transaction': True})
            progress.update(
                step7_start_mariadb, description='[green]MariaDB server started ✓', completed=True
            )
            progress.stop_task(step7_start_mariadb)

            step8_start_cluster = progress.add_task('Starting MCS cluster...', total=None)
            cluster_start_resp = cluster_api_client.start_cluster(
                {'in_transaction': True}
            )
            with TransactionManager(
                timeout=timedelta(days=1).total_seconds(), handle_signals=True
            ):
                cluster_start_resp = cluster_api_client.start_cluster({'in_transaction': True})
            progress.update(
                step8_start_cluster, description='[green]MCS Cluster started ✓', completed=True
            )
            progress.stop_task(step8_start_cluster)
            post_print('Upgrade completed and services restarted successfully.', 'green')

    # Render any deferred output now that the progress bar is complete
    for item in post_output:
        console.print(item)

    raise typer.Exit(code=exit_code)



import sys
import re

"""
Example usage:
    python check_log_for_xa.py path/to/log/mysqld.2.err rollback,commit,prepare
"""

log_path = sys.argv[1]
desired_filters = sys.argv[2]

all_filters = [
  ('rollback', re.compile('(\[Note\] rollback xid .+)')),
  ('commit', re.compile('(\[Note\] commit xid .+)')),
  ('prepare',
    re.compile('(\[Note\] Found \d+ prepared transaction\(s\) in \w+)')),
]

active_filters = filter(lambda f: f[0] in desired_filters, all_filters)

results = set()
with open(log_path) as log:
  for line in log:
    line = line.strip()
    for f in active_filters:
      match = f[1].search(line)
      if match:
        results.add("**found '%s' log entry**" % f[0])

for res in results:
  print res



# Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

from advisor.rule_parser import RulesSpec
from advisor.db_log_parser import DatabaseLogs, DataSource
from advisor.db_options_parser import DatabaseOptions
from advisor.db_stats_fetcher import LogStatsParser, OdsStatsFetcher
import argparse


def main(args):
    # initialise the RulesSpec parser
    rule_spec_parser = RulesSpec(args.rules_spec)
    rule_spec_parser.load_rules_from_spec()
    rule_spec_parser.perform_section_checks()
    # initialize the DatabaseOptions object
    db_options = DatabaseOptions(args.rocksdb_options)
    # Create DatabaseLogs object
    db_logs = DatabaseLogs(
        args.log_files_path_prefix, db_options.get_column_families()
    )
    # Create the Log STATS object
    db_log_stats = LogStatsParser(
        args.log_files_path_prefix, args.stats_dump_period_sec
    )
    data_sources = {
        DataSource.Type.DB_OPTIONS: [db_options],
        DataSource.Type.LOG: [db_logs],
        DataSource.Type.TIME_SERIES: [db_log_stats]
    }
    if args.ods_client:
        data_sources[DataSource.Type.TIME_SERIES].append(OdsStatsFetcher(
            args.ods_client,
            args.ods_entity,
            args.ods_tstart,
            args.ods_tend,
            args.ods_key_prefix
        ))
    triggered_rules = rule_spec_parser.get_triggered_rules(
        data_sources, db_options.get_column_families()
    )
    rule_spec_parser.print_rules(triggered_rules)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Use this script to get\
        suggestions for improving Rocksdb performance.')
    parser.add_argument(
        '--rules_spec', required=True, type=str,
        help='path of the file containing the expert-specified Rules'
    )
    parser.add_argument(
        '--rocksdb_options', required=True, type=str,
        help='path of the starting Rocksdb OPTIONS file'
    )
    parser.add_argument(
        '--log_files_path_prefix', required=True, type=str,
        help='path prefix of the Rocksdb LOG files'
    )
    parser.add_argument(
        '--stats_dump_period_sec', required=True, type=int,
        help='the frequency (in seconds) at which STATISTICS are printed to ' +
        'the Rocksdb LOG file'
    )
    # ODS arguments
    parser.add_argument(
        '--ods_client', type=str, help='the ODS client binary'
    )
    parser.add_argument(
        '--ods_entity', type=str,
        help='the servers for which the ODS stats need to be fetched'
    )
    parser.add_argument(
        '--ods_key_prefix', type=str,
        help='the prefix that needs to be attached to the keys of time ' +
        'series to be fetched from ODS'
    )
    parser.add_argument(
        '--ods_tstart', type=int,
        help='start time of timeseries to be fetched from ODS'
    )
    parser.add_argument(
        '--ods_tend', type=int,
        help='end time of timeseries to be fetched from ODS'
    )
    args = parser.parse_args()
    main(args)



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from __future__ import print_function

import optparse
import re
import sys

# the gcov report follows certain pattern. Each file will have two lines
# of report, from which we can extract the file name, total lines and coverage
# percentage.
def parse_gcov_report(gcov_input):
    per_file_coverage = {}
    total_coverage = None

    for line in sys.stdin:
        line = line.strip()

        # --First line of the coverage report (with file name in it)?
        match_obj = re.match("^File '(.*)'$", line)
        if match_obj:
            # fetch the file name from the first line of the report.
            current_file = match_obj.group(1)
            continue

        # -- Second line of the file report (with coverage percentage)
        match_obj = re.match("^Lines executed:(.*)% of (.*)", line)

        if match_obj:
            coverage = float(match_obj.group(1))
            lines = int(match_obj.group(2))

            if current_file is not None:
                per_file_coverage[current_file] = (coverage, lines)
                current_file = None
            else:
                # If current_file is not set, we reach the last line of report,
                # which contains the summarized coverage percentage.
                total_coverage = (coverage, lines)
            continue

        # If the line's pattern doesn't fall into the above categories. We
        # can simply ignore them since they're either empty line or doesn't
        # find executable lines of the given file.
        current_file = None

    return per_file_coverage, total_coverage

def get_option_parser():
    usage = "Parse the gcov output and generate more human-readable code " +\
            "coverage report."
    parser = optparse.OptionParser(usage)

    parser.add_option(
        "--interested-files", "-i",
        dest="filenames",
        help="Comma separated files names. if specified, we will display " +
             "the coverage report only for interested source files. " +
             "Otherwise we will display the coverage report for all " +
             "source files."
    )
    return parser

def display_file_coverage(per_file_coverage, total_coverage):
    # To print out auto-adjustable column, we need to know the longest
    # length of file names.
    max_file_name_length = max(
        len(fname) for fname in per_file_coverage.keys()
    )

    # -- Print header
    # size of separator is determined by 3 column sizes:
    # file name, coverage percentage and lines.
    header_template = \
        "%" + str(max_file_name_length) + "s\t%s\t%s"
    separator = "-" * (max_file_name_length + 10 + 20)
    print(header_template % ("Filename", "Coverage", "Lines"))  # noqa: E999 T25377293 Grandfathered in
    print(separator)

    # -- Print body
    # template for printing coverage report for each file.
    record_template = "%" + str(max_file_name_length) + "s\t%5.2f%%\t%10d"

    for fname, coverage_info in per_file_coverage.items():
        coverage, lines = coverage_info
        print(record_template % (fname, coverage, lines))

    # -- Print footer
    if total_coverage:
        print(separator)
        print(record_template % ("Total", total_coverage[0], total_coverage[1]))

def report_coverage():
    parser = get_option_parser()
    (options, args) = parser.parse_args()

    interested_files = set()
    if options.filenames is not None:
        interested_files = set(f.strip() for f in options.filenames.split(','))

    # To make things simple, right now we only read gcov report from the input
    per_file_coverage, total_coverage = parse_gcov_report(sys.stdin)

    # Check if we need to display coverage info for interested files.
    if len(interested_files):
        per_file_coverage = dict(
            (fname, per_file_coverage[fname]) for fname in interested_files
            if fname in per_file_coverage
        )
        # If we only interested in several files, it makes no sense to report
        # the total_coverage
        total_coverage = None

    if not len(per_file_coverage):
        print("Cannot find coverage info for the given files.", file=sys.stderr)
        return
    display_file_coverage(per_file_coverage, total_coverage)

if __name__ == "__main__":
    report_coverage()



# Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

import copy
from advisor.db_log_parser import DataSource, NO_COL_FAMILY
from advisor.ini_parser import IniParser
import os


class OptionsSpecParser(IniParser):
    @staticmethod
    def is_new_option(line):
        return '=' in line

    @staticmethod
    def get_section_type(line):
        '''
        Example section header: [TableOptions/BlockBasedTable "default"]
        Here ConfigurationOptimizer returned would be
        'TableOptions.BlockBasedTable'
        '''
        section_path = line.strip()[1:-1].split()[0]
        section_type = '.'.join(section_path.split('/'))
        return section_type

    @staticmethod
    def get_section_name(line):
        # example: get_section_name('[CFOptions "default"]')
        token_list = line.strip()[1:-1].split('"')
        # token_list = ['CFOptions', 'default', '']
        if len(token_list) < 3:
            return None
        return token_list[1]  # return 'default'

    @staticmethod
    def get_section_str(section_type, section_name):
        # Example:
        # Case 1: get_section_str('DBOptions', NO_COL_FAMILY)
        # Case 2: get_section_str('TableOptions.BlockBasedTable', 'default')
        section_type = '/'.join(section_type.strip().split('.'))
        # Case 1: section_type = 'DBOptions'
        # Case 2: section_type = 'TableOptions/BlockBasedTable'
        section_str = '[' + section_type
        if section_name == NO_COL_FAMILY:
            # Case 1: '[DBOptions]'
            return (section_str + ']')
        else:
            # Case 2: '[TableOptions/BlockBasedTable "default"]'
            return section_str + ' "' + section_name + '"]'

    @staticmethod
    def get_option_str(key, values):
        option_str = key + '='
        # get_option_str('db_log_dir', None), returns 'db_log_dir='
        if values:
            # example:
            # get_option_str('max_bytes_for_level_multiplier_additional',
            # [1,1,1,1,1,1,1]), returned string:
            # 'max_bytes_for_level_multiplier_additional=1:1:1:1:1:1:1'
            if isinstance(values, list):
                for value in values:
                    option_str += (str(value) + ':')
                option_str = option_str[:-1]
            else:
                # example: get_option_str('write_buffer_size', 1048576)
                # returned string: 'write_buffer_size=1048576'
                option_str += str(values)
        return option_str


class DatabaseOptions(DataSource):

    @staticmethod
    def is_misc_option(option_name):
        # these are miscellaneous options that are not yet supported by the
        # Rocksdb options file, hence they are not prefixed with any section
        # name
        return '.' not in option_name

    @staticmethod
    def get_options_diff(opt_old, opt_new):
        # type: Dict[option, Dict[col_fam, value]] X 2 ->
        # Dict[option, Dict[col_fam, Tuple(old_value, new_value)]]
        # note: diff should contain a tuple of values only if they are
        # different from each other
        options_union = set(opt_old.keys()).union(set(opt_new.keys()))
        diff = {}
        for opt in options_union:
            diff[opt] = {}
            # if option in options_union, then it must be in one of the configs
            if opt not in opt_old:
                for col_fam in opt_new[opt]:
                    diff[opt][col_fam] = (None, opt_new[opt][col_fam])
            elif opt not in opt_new:
                for col_fam in opt_old[opt]:
                    diff[opt][col_fam] = (opt_old[opt][col_fam], None)
            else:
                for col_fam in opt_old[opt]:
                    if col_fam in opt_new[opt]:
                        if opt_old[opt][col_fam] != opt_new[opt][col_fam]:
                            diff[opt][col_fam] = (
                                opt_old[opt][col_fam],
                                opt_new[opt][col_fam]
                            )
                    else:
                        diff[opt][col_fam] = (opt_old[opt][col_fam], None)
                for col_fam in opt_new[opt]:
                    if col_fam in opt_old[opt]:
                        if opt_old[opt][col_fam] != opt_new[opt][col_fam]:
                            diff[opt][col_fam] = (
                                opt_old[opt][col_fam],
                                opt_new[opt][col_fam]
                            )
                    else:
                        diff[opt][col_fam] = (None, opt_new[opt][col_fam])
            if not diff[opt]:
                diff.pop(opt)
        return diff

    def __init__(self, rocksdb_options, misc_options=None):
        super().__init__(DataSource.Type.DB_OPTIONS)
        # The options are stored in the following data structure:
        # Dict[section_type, Dict[section_name, Dict[option_name, value]]]
        self.options_dict = None
        self.column_families = None
        # Load the options from the given file to a dictionary.
        self.load_from_source(rocksdb_options)
        # Setup the miscellaneous options expected to be List[str], where each
        # element in the List has the format "<option_name>=<option_value>"
        # These options are the ones that are not yet supported by the Rocksdb
        # OPTIONS file, so they are provided separately
        self.setup_misc_options(misc_options)

    def setup_misc_options(self, misc_options):
        self.misc_options = {}
        if misc_options:
            for option_pair_str in misc_options:
                option_name = option_pair_str.split('=')[0].strip()
                option_value = option_pair_str.split('=')[1].strip()
                self.misc_options[option_name] = option_value

    def load_from_source(self, options_path):
        self.options_dict = {}
        with open(options_path, 'r') as db_options:
            for line in db_options:
                line = OptionsSpecParser.remove_trailing_comment(line)
                if not line:
                    continue
                if OptionsSpecParser.is_section_header(line):
                    curr_sec_type = (
                        OptionsSpecParser.get_section_type(line)
                    )
                    curr_sec_name = OptionsSpecParser.get_section_name(line)
                    if curr_sec_type not in self.options_dict:
                        self.options_dict[curr_sec_type] = {}
                    if not curr_sec_name:
                        curr_sec_name = NO_COL_FAMILY
                    self.options_dict[curr_sec_type][curr_sec_name] = {}
                    # example: if the line read from the Rocksdb OPTIONS file
                    # is [CFOptions "default"], then the section type is
                    # CFOptions and 'default' is the name of a column family
                    # that for this database, so it's added to the list of
                    # column families stored in this object
                    if curr_sec_type == 'CFOptions':
                        if not self.column_families:
                            self.column_families = []
                        self.column_families.append(curr_sec_name)
                elif OptionsSpecParser.is_new_option(line):
                    key, value = OptionsSpecParser.get_key_value_pair(line)
                    self.options_dict[curr_sec_type][curr_sec_name][key] = (
                        value
                    )
                else:
                    error = 'Not able to parse line in Options file.'
                    OptionsSpecParser.exit_with_parse_error(line, error)

    def get_misc_options(self):
        # these are options that are not yet supported by the Rocksdb OPTIONS
        # file, hence they are provided and stored separately
        return self.misc_options

    def get_column_families(self):
        return self.column_families

    def get_all_options(self):
        # This method returns all the options that are stored in this object as
        # a: Dict[<sec_type>.<option_name>: Dict[col_fam, option_value]]
        all_options = []
        # Example: in the section header '[CFOptions "default"]' read from the
        # OPTIONS file, sec_type='CFOptions'
        for sec_type in self.options_dict:
            for col_fam in self.options_dict[sec_type]:
                for opt_name in self.options_dict[sec_type][col_fam]:
                    option = sec_type + '.' + opt_name
                    all_options.append(option)
        all_options.extend(list(self.misc_options.keys()))
        return self.get_options(all_options)

    def get_options(self, reqd_options):
        # type: List[str] -> Dict[str, Dict[str, Any]]
        # List[option] -> Dict[option, Dict[col_fam, value]]
        reqd_options_dict = {}
        for option in reqd_options:
            if DatabaseOptions.is_misc_option(option):
                # the option is not prefixed by '<section_type>.' because it is
                # not yet supported by the Rocksdb OPTIONS file; so it has to
                # be fetched from the misc_options dictionary
                if option not in self.misc_options:
                    continue
                if option not in reqd_options_dict:
                    reqd_options_dict[option] = {}
                reqd_options_dict[option][NO_COL_FAMILY] = (
                    self.misc_options[option]
                )
            else:
                # Example: option = 'TableOptions.BlockBasedTable.block_align'
                # then, sec_type = 'TableOptions.BlockBasedTable'
                sec_type = '.'.join(option.split('.')[:-1])
                # opt_name = 'block_align'
                opt_name = option.split('.')[-1]
                if sec_type not in self.options_dict:
                    continue
                for col_fam in self.options_dict[sec_type]:
                    if opt_name in self.options_dict[sec_type][col_fam]:
                        if option not in reqd_options_dict:
                            reqd_options_dict[option] = {}
                        reqd_options_dict[option][col_fam] = (
                            self.options_dict[sec_type][col_fam][opt_name]
                        )
        return reqd_options_dict

    def update_options(self, options):
        # An example 'options' object looks like:
        # {'DBOptions.max_background_jobs': {NO_COL_FAMILY: 2},
        # 'CFOptions.write_buffer_size': {'default': 1048576, 'cf_A': 128000},
        # 'bloom_bits': {NO_COL_FAMILY: 4}}
        for option in options:
            if DatabaseOptions.is_misc_option(option):
                # this is a misc_option i.e. an option that is not yet
                # supported by the Rocksdb OPTIONS file, so it is not prefixed
                # by '<section_type>.' and must be stored in the separate
                # misc_options dictionary
                if NO_COL_FAMILY not in options[option]:
                    print(
                        'WARNING(DatabaseOptions.update_options): not ' +
                        'updating option ' + option + ' because it is in ' +
                        'misc_option format but its scope is not ' +
                        NO_COL_FAMILY + '. Check format of option.'
                    )
                    continue
                self.misc_options[option] = options[option][NO_COL_FAMILY]
            else:
                sec_name = '.'.join(option.split('.')[:-1])
                opt_name = option.split('.')[-1]
                if sec_name not in self.options_dict:
                    self.options_dict[sec_name] = {}
                for col_fam in options[option]:
                    # if the option is not already present in the dictionary,
                    # it will be inserted, else it will be updated to the new
                    # value
                    if col_fam not in self.options_dict[sec_name]:
                        self.options_dict[sec_name][col_fam] = {}
                    self.options_dict[sec_name][col_fam][opt_name] = (
                        copy.deepcopy(options[option][col_fam])
                    )

    def generate_options_config(self, nonce):
        # this method generates a Rocksdb OPTIONS file in the INI format from
        # the options stored in self.options_dict
        this_path = os.path.abspath(os.path.dirname(__file__))
        file_name = '../temp/OPTIONS_' + str(nonce) + '.tmp'
        file_path = os.path.join(this_path, file_name)
        with open(file_path, 'w') as fp:
            for section in self.options_dict:
                for col_fam in self.options_dict[section]:
                    fp.write(
                        OptionsSpecParser.get_section_str(section, col_fam) +
                        '\n'
                    )
                    for option in self.options_dict[section][col_fam]:
                        values = self.options_dict[section][col_fam][option]
                        fp.write(
                            OptionsSpecParser.get_option_str(option, values) +
                            '\n'
                        )
                fp.write('\n')
        return file_path

    def check_and_trigger_conditions(self, conditions):
        for cond in conditions:
            reqd_options_dict = self.get_options(cond.options)
            # This contains the indices of options that are specific to some
            # column family and are not database-wide options.
            incomplete_option_ix = []
            options = []
            missing_reqd_option = False
            for ix, option in enumerate(cond.options):
                if option not in reqd_options_dict:
                    print(
                        'WARNING(DatabaseOptions.check_and_trigger): ' +
                        'skipping condition ' + cond.name + ' because it '
                        'requires option ' + option + ' but this option is' +
                        ' not available'
                    )
                    missing_reqd_option = True
                    break  # required option is absent
                if NO_COL_FAMILY in reqd_options_dict[option]:
                    options.append(reqd_options_dict[option][NO_COL_FAMILY])
                else:
                    options.append(None)
                    incomplete_option_ix.append(ix)

            if missing_reqd_option:
                continue

            # if all the options are database-wide options
            if not incomplete_option_ix:
                try:
                    if eval(cond.eval_expr):
                        cond.set_trigger({NO_COL_FAMILY: options})
                except Exception as e:
                    print(
                        'WARNING(DatabaseOptions) check_and_trigger:' + str(e)
                    )
                continue

            # for all the options that are not database-wide, we look for their
            # values specific to column families
            col_fam_options_dict = {}
            for col_fam in self.column_families:
                present = True
                for ix in incomplete_option_ix:
                    option = cond.options[ix]
                    if col_fam not in reqd_options_dict[option]:
                        present = False
                        break
                    options[ix] = reqd_options_dict[option][col_fam]
                if present:
                    try:
                        if eval(cond.eval_expr):
                            col_fam_options_dict[col_fam] = (
                                copy.deepcopy(options)
                            )
                    except Exception as e:
                        print(
                            'WARNING(DatabaseOptions) check_and_trigger: ' +
                            str(e)
                        )
            # Trigger for an OptionCondition object is of the form:
            # Dict[col_fam_name: List[option_value]]
            # where col_fam_name is the name of a column family for which
            # 'eval_expr' evaluated to True and List[option_value] is the list
            # of values of the options specified in the condition's 'options'
            # field
            if col_fam_options_dict:
                cond.set_trigger(col_fam_options_dict)



"""Module contains custom exceptions."""

from collections.abc import Iterator
from contextlib import contextmanager
from typing import Optional, TypeVar, Any

from pydantic import BaseModel, ValidationError

from cmapi_server.controllers.error import APIError


class CMAPIBasicError(Exception):
    """Basic exception raised for CMAPI related processes.

    Attributes:
        message -- explanation of the error
    """
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)
    def __str__(self) -> str:
        return self.message


class CEJError(CMAPIBasicError):
    """Exception raised for CEJ related processes.

    Attributes:
        message -- explanation of the error
    """


@contextmanager
def exc_to_cmapi_error(prefix: Optional[str] = None) -> Iterator[None]:
    """Context manager to standardize error wrapping into CMAPIBasicError.

    Re-raises existing CMAPIBasicError untouched (to preserve detailed
    messages). Any other exception type is wrapped into CMAPIBasicError with an
    optional prefix and the original exception string appended as details.

    :param prefix: Optional message prefix for wrapped errors
    :raises CMAPIBasicError: for any wrapped non-CMAPIBasicError exceptions
    """
    try:
        yield
    except CMAPIBasicError:
        # Preserve detailed messages from deeper layers (e.g., validation)
        raise
    except Exception as err:
        msg = f"{prefix}. Details: {err}" if prefix else str(err)
        raise CMAPIBasicError(msg) from err


@contextmanager
def cmapi_error_to_422(logger, func_name: str) -> Iterator[None]:
    """Convert CMAPIBasicError to HTTP 422 APIError."""
    try:
        yield
    except CMAPIBasicError as err:
        # mirror raise_422_error behavior locally to avoid circular imports
        logger.error(f'{func_name} {err.message}', exc_info=False)
        raise APIError(422, err.message) from err


T = TypeVar('T', bound=BaseModel)

def validate_or_422(
    model: type[T], payload: Any, logger, func_name: str,
    prefix: Optional[str] = 'Invalid request body',
) -> T:
    """Validate payload with Pydantic model or raise HTTP 422 APIError."""
    try:
        return model.model_validate(payload)
    except ValidationError as exp:
        msg = f"{prefix}: {exp.errors()}" if prefix else str(exp.errors())
        logger.error(f"{func_name} {msg}", exc_info=False)
        raise APIError(422, msg) from exp


@contextmanager
def exc_to_422(logger, func_name: str, prefix: Optional[str] = None) -> Iterator[None]:
    """Convert any exception into HTTP 422"""
    with cmapi_error_to_422(logger, func_name):
        with exc_to_cmapi_error(prefix=prefix):
            yield



"""
Module contains non-systemd/container process dispatcher class implementation.
"""

import logging
import os.path
import re
from pathlib import Path
from time import sleep

import psutil

from cmapi_server.constants import (
    IFLAG, LIBJEMALLOC_DEFAULT_PATH, MCS_INSTALL_BIN, ALL_MCS_PROGS, MCSProgs
)
from cmapi_server.process_dispatchers.locks import release_shmem_locks
from cmapi_server.exceptions import CMAPIBasicError
from cmapi_server.process_dispatchers.base import BaseDispatcher


class ContainerDispatcher(BaseDispatcher):
    """Manipulates processes in docker container.

    It's possible to use in any OS/container environment in cases when
    we don't want to use systemd or don't have it.
    """
    libjemalloc_path = None

    @staticmethod
    def _set_iflag():
        """Create IFLAG file.

        Means Columnstore container init finished.
        """
        Path(IFLAG).touch()

    @classmethod
    def _get_proc_object(cls, name: str) -> psutil.Process:
        """Getting psutil Process object by service name.

        :param name: process name
        :type name: str
        :raises psutil.NoSuchProcess: if no process with such name presented
        :return: Process object with specified name
        :rtype: psutil.Process

        ...TODO: add types-psutil to requirements for mypy checks
        """
        for proc in psutil.process_iter(['pid', 'name', 'username']):
            if proc.name().lower() == name.lower():
                return proc
        raise psutil.NoSuchProcess(pid=None, name=name)

    @classmethod
    def get_libjemalloc_path(cls) -> str:
        """Get libjemalloc.so path.

        :raises CMAPIBasicError: raises if ldconfig execution returned non zero
        :raises FileNotFoundError: if no libjemalloc.so.2 found
        :return: libjemalloc.so.2 path
        :rtype: str
        """
        logger = logging.getLogger('container_sh')
        if cls.libjemalloc_path:
            return cls.libjemalloc_path
        # pylint: disable=line-too-long
        # for reference: https://github.com/pyinstaller/pyinstaller/blob/f29b577df4e1659cf65aacb797034763308fd298/PyInstaller/depend/utils.py#L304

        splitlines_count = 1
        pattern = re.compile(r'^\s+(\S+)(\s.*)? => (\S+)')
        success, result = cls.exec_command('ldconfig -p')
        if not success:
            raise CMAPIBasicError('Failed executing ldconfig.')

        text = result.strip().splitlines()[splitlines_count:]

        for line in text:
            # this assumes library names do not contain whitespace
            p_match = pattern.match(line)
            # Sanitize away any abnormal lines of output.
            if p_match is None:
                continue

            lib_path = p_match.groups()[-1]
            lib_name = p_match.group(1)
            if 'libjemalloc' in lib_name:
                # use the first entry
                # TODO: do we need path or name here?
                # $(ldconfig -p | grep -m1 libjemalloc | awk '{print $1}')
                cls.libjemalloc_path = lib_path
                break

        if not cls.libjemalloc_path:
            if not os.path.exists(LIBJEMALLOC_DEFAULT_PATH):
                logger.error('No libjemalloc.so.2 found.')
                raise FileNotFoundError
            cls.libjemalloc_path = LIBJEMALLOC_DEFAULT_PATH

        return cls.libjemalloc_path

    @classmethod
    def is_service_running(cls, service: str, use_sudo: bool = True) -> bool:
        """Check if MCS process (not only MCS, but also any other process) is running.

        :param service: service name
        :type service: str
        :param use_sudo: interface requirement, unused here, defaults to True
        :type use_sudo: bool, optional
        :return: True if service is running, otherwise False
        :rtype: bool

        ..Note:
            Not working with multiple services at a time.
        """
        try:
            cls._get_proc_object(service)
        except psutil.NoSuchProcess:
            return False
        return True

    @staticmethod
    def _make_cmd(service: str) -> str:
        """Make shell command by service name.

        :param service: service name
        :type service: str
        :return: command with arguments if needed
        :rtype: str
        """
        prog = MCSProgs(service)
        service_info = ALL_MCS_PROGS[prog]
        command = os.path.join(MCS_INSTALL_BIN, service)

        if service_info.subcommand:
            subcommand = service_info.subcommand
            command = f'{command} {subcommand}'

        return command

    @classmethod
    def start(
        cls, service: str, is_primary: bool, use_sudo: bool = True
    ) -> bool:
        """Start process in docker container.

        :param service: process name
        :type service: str
        :param is_primary: is node primary or not
        :type is_primary: bool, optional
        :param use_sudo: interface required, unused here, defaults to True
        :type use_sudo: bool, optional
        :return: True if service started successfully
        :rtype: bool
        """
        logger = logging.getLogger('container_sh')
        if cls.is_service_running(service):
            return True

        logger.debug(f'Starting {service}')
        env_vars = {"LD_PRELOAD": cls.get_libjemalloc_path()}
        command = cls._make_cmd(service)

        if service == 'workernode':
            # workernode starts on primary and non primary node with 1 or 2
            # added to the end of argument:
            # DBRM_Worker1 - on primary, DBRM_Worker2 - non primary
            command = command.format(1 if is_primary else 2)

            # start mcs-loadbrm.py before workernode
            logger.debug('Waiting to load BRM.')
            loadbrm_path = os.path.join(MCS_INSTALL_BIN, 'mcs-loadbrm.py')
            loadbrm_logpath = cls._create_mcs_process_logfile(
                'mcs-loadbrm.log'
            )
            with open(loadbrm_logpath, 'a', encoding='utf-8') as loadbrm_logfh:
                success, _ = cls.exec_command(
                    f'{loadbrm_path} no', stdout=loadbrm_logfh, env=env_vars
                )
            if not success:
                logger.error('Error while loading BRM.')
            else:
                logger.debug('Successfully loaded BRM.')

        service_log_path = cls._create_mcs_process_logfile(
            f'{service.lower()}.log'
        )
        success, _ = cls.exec_command(
            command, daemonize=True,
            stdout=open(service_log_path, 'a', encoding='utf-8'),
            env=env_vars
        )
        # TODO: any other way to detect service finished its initialisation?
        prog = MCSProgs(service)
        sleep(ALL_MCS_PROGS[prog].delay)
        logger.debug(f'Started "{service}".')

        if is_primary and service == 'DDLProc':
            cls._run_dbbuilder()

        return cls.is_service_running(service)

    @classmethod
    def stop(
        cls, service: str, is_primary: bool, use_sudo: bool = True
    ) -> bool:
        """Stop process in docker container.

        :param service: process name
        :type service: str
        :param is_primary: is node primary or not
        :type is_primary: bool, optional
        :param use_sudo: interface required, unused here, defaults to True
        :type use_sudo: bool, optional
        :return: True if service started successfully
        :rtype: bool
        """
        logger = logging.getLogger('container_sh')
        if not cls.is_service_running(service):
            return True

        logger.debug(f'Stopping {service}')
        service_proc = cls._get_proc_object(service)

        if service == 'workernode':
            # Run pre-stop lock reset before saving BRM
            # These stale locks can occur if the controllernode couldn't stop correctly
            #  and they cause mcs-savebrm.py to hang
            release_shmem_locks(logger)

            # start mcs-savebrm.py before stoping workernode
            logger.debug('Waiting to save BRM.')
            savebrm_path = os.path.join(MCS_INSTALL_BIN, 'mcs-savebrm.py')
            savebrm_logpath = cls._create_mcs_process_logfile(
                'mcs-savebrm.log'
            )
            with open(savebrm_logpath, 'a', encoding='utf-8') as savebrm_logfh:
                success, _ = cls.exec_command(
                    savebrm_path, stdout=savebrm_logfh
                )
            if not success:
                logger.error('Error while saving BRM.')
            else:
                logger.debug('Successfully saved BRM.')

            logger.debug('Start clearing SHM.')
            clearshm_path = os.path.join(MCS_INSTALL_BIN, 'clearShm')
            success, _ = cls.exec_command(clearshm_path)
            if not success:
                logger.error('Error while clearing SHM.')
            else:
                logger.debug('Successfully cleared SHM.')

        service_proc.terminate()
        # timeout got from old container.sh
        # TODO: this is still not enough for controllernode process
        #       it should be always stop by SIGKILL, need to investigate.
        timeout = 3
        if service == 'StorageManager':
            timeout = 300  # 5 minutes
        logger.debug(f'Waiting to gracefully stop "{service}".')
        # This function will return as soon as all processes terminate
        # or when timeout (seconds) occurs.
        gone, alive = psutil.wait_procs([service_proc], timeout=timeout)
        if alive:
            logger.debug(
                f'{service} not terminated with SIGTERM, sending SIGKILL.'
            )
            # only one process could be in a list
            alive[0].kill()
            gone, alive = psutil.wait_procs([service_proc], timeout=timeout)
            if gone:
                logger.debug(f'Successfully killed "{service}".')
            else:
                logger.warning(
                    f'Service "{service}" still alive after sending "kill -9" '
                    f'and waiting {timeout} seconds.'
                )
        else:
            logger.debug(f'Gracefully stopped "{service}".')

        return not cls.is_service_running(service)

    @classmethod
    def restart(
        cls, service: str, is_primary: bool, use_sudo: bool = True
    ) -> bool:
        """Restart process in docker container.

        :param service: process name
        :type service: str
        :param is_primary: is node primary or not
        :type is_primary: bool, optional
        :param use_sudo: interface required, unused here, defaults to True
        :type use_sudo: bool, optional
        :return: True if service started successfully
        :rtype: bool

        ...TODO: for next releases. Additional error handling.
        """
        stop_success = True
        if cls.is_service_running(service):
            # TODO: retry?
            stop_success = cls.stop(service, is_primary, use_sudo)
        start_success = cls.start(service, is_primary, use_sudo)

        return stop_success and start_success



from integration_tests.ssh import ClusterConfig
from integration_tests.utils import assert_dict_includes, drop_timestamp


class TestRemoveNode:
    def test_remove_node(self, cluster_config: ClusterConfig, complete_cluster):
        """Test removing hosts from the cluster"""
        # Remove replicas
        remove_cmd = "cluster node remove "
        for replica in cluster_config.replicas:
            remove_cmd += f"--node {replica.private_ip} "
        remove_out = cluster_config.primary.exec_mcs(remove_cmd)
        assert len(remove_out) == len(cluster_config.replicas)
        removed_node_ids = {node["node_id"] for node in remove_out}
        expected_node_ids = {str(replica.private_ip) for replica in cluster_config.replicas}
        assert removed_node_ids == expected_node_ids

        status_out = cluster_config.primary.exec_mcs("cluster status")
        assert status_out["num_nodes"] == 1
        assert str(cluster_config.primary.private_ip) in status_out

        primary_status = drop_timestamp(status_out[str(cluster_config.primary.private_ip)])
        assert_dict_includes(
            primary_status,
            {
                "dbrm_mode": "master",
                "cluster_mode": "readwrite",
                "dbroots": ['1', '2', '3'],  # All the dbroots are rebalanced to the last host
                "module_id": 1,
            },
        )

        # Remove primary node
        print(f"Removing primary node ({cluster_config.primary}) from the cluster")
        remove_out = cluster_config.primary.exec_mcs(f"cluster node remove --node {cluster_config.primary.private_ip}")
        assert len(remove_out) == 1
        assert remove_out[0]["node_id"] == str(cluster_config.primary.private_ip)

        status_out = cluster_config.primary.exec_mcs("cluster status")
        print(status_out)
        assert status_out["num_nodes"] == 0
        assert str(cluster_config.primary.private_ip) not in status_out



import re
from dataclasses import dataclass
from functools import total_ordering


@total_ordering
@dataclass
class ComparableVersion:
    version: str

    def __post_init__(self):
        self.version_nums = self._split_version(self.version)

    def _split_version(self, version: str) -> list[int]:
        # Drop epoch if present (debian style)
        if ':' in version:
            version = version.split(':', 1)[1]

        # I'm not sure if ~ could be the first, but seems to be possible in some debians
        # Trim at first suffix marker (+, ~) if any
        for marker in ('+', '~'):
            if marker in version:
                version = version.split(marker, 1)[0]
                break
        # Remove leading .0s and split by . _ -
        parts = re.split(r'[._-]', version)
        return [int(p.lstrip('0') or '0') for p in parts]

    def _normalized(self, other) -> tuple[list, list]:
        """Return two zero-padded version lists of equal length."""
        v1 = self.version_nums.copy()
        v2 = other.version_nums.copy()
        max_len = max(len(v1), len(v2))
        v1 += [0] * (max_len - len(v1))
        v2 += [0] * (max_len - len(v2))
        return v1, v2

    def __eq__(self, other):
        v1, v2 = self._normalized(other)
        return v1 == v2

    def __lt__(self, other):
        v1, v2 = self._normalized(other)
        return v1 < v2
